import asyncio
import logging
import os
import random
import time

import aiohttp
import hydra
from hydra.utils import instantiate
from omegaconf import DictConfig

from examples.rl_webagent.agent import WebAgent
from examples.rl_webagent.scripts.orchestrate_rl import load_webtasks
from examples.rl_webagent.steps import WebTape
from tapeagents.core import Observation
from tapeagents.io import save_json_tape
from tapeagents.orchestrator import async_execute_agent
from tapeagents.remote_environment import AsyncRemoteEnvironment

logger = logging.getLogger(__name__)


def abt_to_json(tasks: list[dict]) -> list[dict]:
    """
    Convert AbstractBrowserTask tasks to dicts that are JSON serializable by only giving the dataset, task_id, and seed.
    """
    return [{"dataset": task["dataset"], "task": task["task"].get_task_id(), "seed": task["seed"]} for task in tasks]


async def run_agent_with_remote_env(
    cfg: DictConfig, task: dict, session: aiohttp.ClientSession, max_loops: int
) -> WebTape:
    environment: AsyncRemoteEnvironment = instantiate(cfg.environment)  # type: ignore
    async with environment.acontext(session, wait_for_env=True) as env:
        start_attempts = cfg.start_attempts
        t = time.perf_counter()
        while True:
            try:
                tape_dict, _ = await env.start_task(task)
                break
            except Exception as e:
                start_attempts -= 1
                if start_attempts <= 0:
                    raise e
                logger.warning(f"Failed to start task, retry after 5 seconds: {e}")
                await asyncio.sleep(5)
        start_time = time.perf_counter() - t
        logger.info(f"Task {task['task']}/{task['seed']} started in {start_time:.2f} seconds")
        tape: WebTape = WebTape(**tape_dict)  # convert http response dict to WebTape object
        t = time.perf_counter()
        try:
            actions = await env.a_actions()
            tools_description = await env.a_tools_description()
            llms = None
            try:
                # try to load multiple LLMs from the configuration and choose one randomly
                llms = instantiate(cfg.llms)
                logger.info(f"Loaded {len(llms)} LLMs from configuration.")
                llm = random.choice(llms)
            except Exception:
                pass
            logger.info(f"Using LLM: {llm.base_url}")
            agent: WebAgent = instantiate(cfg.agent, known_actions=actions, tools_description=tools_description)
            agent.llms["default"] = llm
            tape = await async_execute_agent(agent, tape, env, session, max_loops=max_loops)
        except Exception as e:
            logger.error(f"Error occurred while running agent: {e}")
            tape.metadata.error = str(e)
        tape.metadata.result = {"execution_time": time.perf_counter() - t, "start_time": start_time}
    # save the tape as we go
    save_json_tape(tape, os.path.join(cfg.exp_path, "tapes"), tape.metadata.parent_id)
    return tape


async def amain(cfg: DictConfig) -> None:
    os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(cfg.exp_path, "tapedata.sqlite")
    os.environ["MINIWOB_URL"] = cfg.environment_variables.miniwob_url
    # os.environ["SNOW_INSTANCE_URL"] = cfg.environment_variables.snow_instance_url
    # os.environ["SNOW_INSTANCE_UNAME"] = cfg.environment_variables.snow_instance_uname
    # os.environ["SNOW_INSTANCE_PWD"] = cfg.environment_variables.snow_instance_pwd

    ### Step 1: load datasets ###
    train_samples, test_samples = load_webtasks(train_split=cfg.train_split, seeds=cfg.seeds)
    # train_samples, test_samples = load_webtasks_debug()  # TODO: load all tasks when ready
    # Convert AbstractBrowserTask to json parsable dicts
    train_samples = abt_to_json(train_samples)
    test_samples = abt_to_json(test_samples)
    samples = test_samples
    dt = time.perf_counter()
    timeout = cfg.requests_timeout
    timeout = aiohttp.ClientTimeout(total=timeout, connect=timeout, sock_read=timeout)
    coroutines = []
    results = []

    async with aiohttp.ClientSession(timeout=timeout) as session:
        for task in samples:
            logging.info(f"Schedule task {task['task']} with seed {task['seed']}")
            coroutines.append(run_agent_with_remote_env(cfg, task, session, max_loops=cfg.max_loops))
        logger.info(f"Solving {len(coroutines)} tasks")
        results: list[WebTape] = await asyncio.gather(*coroutines)

    logger.info(f"Saved {len(results)} tapes to {os.path.join(cfg.exp_path, 'tapes')}")

    ### Print some statistics
    total_steps = 0
    errs = 0
    no_reward = 0
    acc = []
    rewards = []
    for tape in results:
        total_steps += len(tape.steps)
        last_obs = [step for step in tape if isinstance(step, Observation)][-1]
        success = last_obs.metadata.other.get("reward", 0.0) > 0.5
        acc.append(success)
        if tape.metadata.error:
            errs += 1
            logger.warning(f"Error in tape {tape.metadata.id}: {tape.metadata.error}")
        elif (
            "info" in last_obs.metadata.other
            and "task_info" in last_obs.metadata.other["info"]
            and "REWARD_GLOBAL" in last_obs.metadata.other["info"]["task_info"]
        ):
            rewards.append(last_obs.metadata.other["info"]["task_info"]["REWARD_GLOBAL"])
        else:
            no_reward += 1
            logger.warning(f"No reward found in last observation of tape {tape.metadata.id}")

    logger.info(f"Average tape length: {(total_steps / len(results)):.2f}")
    logger.info(f"Total execution time: {time.perf_counter() - dt:.2f} seconds")
    logger.info(f"Average time per tape: {(time.perf_counter() - dt) / len(results):.2f} seconds")
    logger.info(f"Accuracy: {sum(acc) / len(acc) if acc else 0:.2f}")
    logger.info(f"Average reward: {sum(rewards) / len(rewards) if rewards else 0:.2f}")
    logger.info(f"Number of tapes with no reward: {no_reward}")
    logger.info(f"Failed tapes: {errs}")

    ### TODO: continue to copy things from orchestrate_rl.py / switch to pipelinerl


@hydra.main(
    version_base=None,
    config_path=".",
    config_name="client_main_loop",
)
def main(cfg: DictConfig) -> None:
    asyncio.run(amain(cfg))


if __name__ == "__main__":
    print("Make sure that you run environment server first using `uv run examples/rl_webagent/async/env_server.py`")
    main()



import logging
import os
import time
from typing import Any, Literal

from browsergym.core.task import AbstractBrowserTask
from browsergym.miniwob import ALL_MINIWOB_TASKS
from browsergym.miniwob.base import AbstractMiniwobTask
from joblib import Parallel, delayed

from tapeagents.core import Action, FinalObservation, LLMOutputParsingFailureAction, Observation
from tapeagents.environment import Environment
from tapeagents.steps import ActionExecutionFailure
from tapeagents.tools.browser import Browser
from tapeagents.utils import FatalError

from .steps import (
    FinalAnswerAction,
    ReflectionThought,
    WebTape,
    WebTapeMetadata,
    WebTask,
)

logger = logging.getLogger(__name__)


class WebEnvironment(Environment):
    """
    Web environment for running tasks.
    Translates action steps into gym browser python commands in the form of a string.
    """

    def __init__(
        self,
        exp_path: str | None,
        headless: bool = True,
        observation_format: Literal["axtree", "html", "markdown_html"] = "html",
    ) -> None:
        super().__init__()
        if exp_path:
            os.makedirs(exp_path, exist_ok=True)
        self.exp_path = exp_path
        self.headless = headless
        self.observation_format = observation_format

    def initialize(self):
        self.timers = {}  # reset timers
        self.browser = Browser(
            headless=self.headless,
            exp_path=self.exp_path,
            mock=True,
            observation_format=self.observation_format,  # type: ignore
            timeout_ms=5000,  # 5 seconds timeout for browser actions
        )

    def start_task(self, task_data: dict) -> tuple[WebTape, dict[str, Any]]:
        # if task_data["task"] is a string, it is the task_id of a registered task
        if isinstance(task_data["task"], str):
            # reconstruct the task entrypoint from the dataset and task_id
            assert "dataset" in task_data, "Task data must contain 'dataset' key."
            if "miniwob" in task_data["dataset"]:
                task_entrypoint = [t for t in ALL_MINIWOB_TASKS if t.get_task_id() == task_data["task"]][0]
            else:
                raise ValueError(f"Unknown dataset {task_data['dataset']} for task {task_data['task']}.")
        else:
            task_entrypoint: AbstractBrowserTask = task_data["task"]
        assert issubclass(task_entrypoint, AbstractBrowserTask)
        seed: int = task_data.get("seed", 42)

        self.timers = {}  # reset timers
        start_start_task = time.perf_counter()
        task_id = f"browsergym/{task_entrypoint.get_task_id()}"
        # logger.info(f"WebEnv.start_task {task_id} starting task...")
        _zero = time.perf_counter()
        info = self.browser.start_task(task_id, seed, wait_for_user_message=False)  # type: ignore
        zero = time.perf_counter() - _zero
        logger.info(f"WebEnv.start_task {task_id} browser.start_task took {zero:.2f}s")
        _one = time.perf_counter()
        obs = self.browser.run_browser_action("noop()")
        one = time.perf_counter() - _one
        logger.info(f"WebEnv.start_task {task_id} browser.run_browser_action took {one:.2f}s")
        _two = time.perf_counter()
        tape = WebTape(
            metadata=WebTapeMetadata(task_name=task_entrypoint.get_task_id(), seed=seed),
            steps=[obs, WebTask(task=info["goal"])],
        )
        two = time.perf_counter() - _two
        logger.info(f"WebEnv.start_task {task_id} WebTape creation took {two:.2f}s")
        self.timers["start_task"] = time.perf_counter() - start_start_task
        logger.info(f"total start_task {task_id}: {self.timers['start_task']:.2f}s")
        return tape, info

    def finish_task(self) -> None:
        start_finish_task = time.perf_counter()
        try:
            self.browser.close()
        except Exception as e:
            logger.exception(f"Failed to properly close task: {e}", stack_info=True)
        self.timers["finish_task"] = time.perf_counter() - start_finish_task

    def validate_task(self, tape: WebTape) -> tuple[bool, dict]:
        start_validate_task = time.perf_counter()
        answer = tape.steps[-1].text if isinstance(tape.steps[-1], FinalAnswerAction) else "Task finished"
        self.browser._env.unwrapped.chat.add_message(role="assistant", msg=answer)
        assert self.browser._env.unwrapped.task is not None
        try:
            reward, stop, message, info = self.browser._env.unwrapped.task.validate(
                self.browser._env.unwrapped.page, self.browser._env.unwrapped.chat.messages
            )
        except Exception as e:
            logger.exception(f"Error during task validation: {e}")
            reward = 0
            stop = True
            message = f"Task validation failed with error: {e}"
            info = {}
        # in AbstractMiniwobTask.validate() the reward is defined as float(info["RAW_REWARD_GLOBAL"] > 0) so it will alwayd be 0 or 1
        # let's keep the raw reward for now and define success as reward > 0.5
        if isinstance(self.browser._env.unwrapped.task, AbstractMiniwobTask) and "RAW_REWARD_GLOBAL" in info:
            reward = info["RAW_REWARD_GLOBAL"]
        result_dict = {
            "reward": reward,
            "stop": stop,
            "message": message,
            "info": info,
        }
        self.timers["validate_task"] = time.perf_counter() - start_validate_task
        return bool(reward > 0.5), result_dict

    def react(self, tape: WebTape) -> WebTape:
        if "react" not in self.timers:
            self.timers["react"] = []
        start_react = time.perf_counter()

        actions: list[Action] = []
        for step in tape.steps[-tape.metadata.n_added_steps :]:
            if isinstance(step, Action):
                actions.append(step)
            elif isinstance(step, ReflectionThought):
                # send reflection to chat for user to see
                self.browser._env.unwrapped.chat.add_message(
                    role="assistant", msg=f"{step.last_action_achieved_effect}\nTodo: {step.next_action}"
                )
        for action in actions:
            try:
                if isinstance(action, LLMOutputParsingFailureAction):
                    continue
                observation = self.step(action)
                if isinstance(observation, ActionExecutionFailure):
                    logger.exception(f"Error during action execution: {observation.error}")
                    tape = tape.append(observation)
                    break
                tape = tape.append(observation)  # type: ignore
            except FatalError:
                self.timers["react"].append(time.perf_counter() - start_react)
                raise
            except Exception as e:
                logger.exception(f"Error during action execution {action}: {e}", stack_info=True)
                tape = tape.append(ActionExecutionFailure(error=str(e)))
                break
        self.timers["react"].append(time.perf_counter() - start_react)
        return tape
        # TODO: MAYBE make sure to update parent_id, author_name, etc... in the new tape.metadata just like in agent.run()

    def react_batch(self, tapes: list[WebTape], n_processes: int) -> list[WebTape]:
        results = Parallel(n_jobs=n_processes)([delayed(self.react)(tape) for tape in tapes])
        return results

    def step(self, action: Action) -> Observation:
        obs = self.browser.run(action)
        if obs.metadata.other.get("env_finished", False):
            obs = FinalObservation(metadata=obs.metadata)
        return obs

    def actions(self) -> tuple[type[Action], ...]:
        return self.browser.actions

    def reset(self):
        try:
            self.browser.reset()
        except Exception as e:
            logger.warning(f"Failed to reset browser: {e}, recreate browser instance instead.")
            try:
                self.browser.close()
            except Exception:
                pass
            self.browser = Browser(
                headless=self.headless,
                exp_path=self.exp_path,
                mock=True,
                observation_format=self.observation_format,  # type: ignore
            )



import asyncio
import json
import logging
import os
import pprint
import time
from typing import Any, Generator

import aiohttp
import litellm
import requests
import transformers
from omegaconf import DictConfig, OmegaConf
from pydantic import Field
from tenacity import retry, stop_after_attempt, wait_exponential

from tapeagents.core import LLMOutput, Prompt, TokenLogprob, TrainingText
from tapeagents.llms.base import LLMEvent, LLMStream
from tapeagents.llms.cached import CachedLLM
from tapeagents.utils import get_step_schemas_from_union_type

# force replacement of the tokenizer during testing
_MOCK_TOKENIZER: str = ""
TAPEAGENTS_LLM_TOKEN = "TAPEAGENTS_LLM_TOKEN"
logger = logging.getLogger(__name__)


def trainable_llm_make_training_text(prompt: Prompt, output: LLMOutput, tokenizer) -> TrainingText:
    """
    Generates training text for LLM fine-tuning by combining prompt and output using tokenizer's chat template.

    Args:
        prompt (Prompt): The input prompt containing conversation messages.
        output (LLMOutput): The model's output/response.
        tokenizer (PreTrainedTokenizer): The tokenizer used to format the conversation.

    Returns:
        TrainingText: A dataclass containing:

            - text (str): The formatted conversation text (prompt + output)
            - n_predicted (int): Length of the output text portion

    Note:
        - Uses tokenizer's chat template to format conversations
        - Removes BOS token if present in the beginning of the text
    """
    prompt_text = tokenizer.apply_chat_template(
        conversation=prompt.messages, tokenize=False, add_generation_prompt=True
    )
    text = tokenizer.apply_chat_template(
        prompt.messages + [{"role": "assistant", "content": output.content}],
        tokenize=False,
    )
    output_text = text[len(prompt_text) :]

    if tokenizer.bos_token and text.startswith(tokenizer.bos_token):
        text = text[len(tokenizer.bos_token) :]

    return TrainingText(text=text, n_predicted=len(output_text))


class TrainableLLM(CachedLLM):
    """
    Class for interacting with trainable language models through OpenAI-compatible API endpoints.

    This class implements functionality for both inference and training-related operations with
    language models served via Text Generation Inference (TGI) or vLLM endpoints that expose
    an OpenAI-compatible API interface. It supports streaming non-streaming and async modes
    and includes methods for token counting and log probability calculations.

    Attributes:
        base_url (str): Base URL of the API endpoint
        api_token (str): Authentication token for API access
    """

    # TODO: use OpenAI Python client when the certificate issue is resolved.
    # TODO: consider using litellm

    base_url: str = "https://api.openai.com"
    api_token: str = Field(default="", exclude=True)
    collect_logprobs: bool = False
    use_litellm_tokenizer_fallback: bool = False
    max_parallel_requests: int = 32
    max_retries: int = 5
    base_delay: float = 0.5
    _semaphore: asyncio.Semaphore

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.api_token = os.getenv(TAPEAGENTS_LLM_TOKEN, "") or os.getenv("OPENAI_API_KEY", "")
        self._semaphore = asyncio.Semaphore(self.max_parallel_requests)

    def get_base_url(self) -> str:
        """
        Returns the base URL for the API endpoint.
        """
        return self.base_url.rstrip("/")

    def parse_completion_logprobs(self, completion_logprobs: list[dict]) -> list[TokenLogprob]:
        logprobs = []
        for logprob in completion_logprobs:
            if logprob:
                try:
                    # We assume that the server was launched with --return-tokens-as-token-ids
                    # and that the tokens are provided as: ['token_id:1271', 'token_id:1505', '
                    logprobs.append(
                        TokenLogprob(
                            token_id=int(logprob["token"].split(":")[-1]),
                            logprob=logprob["logprob"],
                        )
                    )
                except Exception as e:
                    logger.error(f"Failed to process logprobs: {logprob}")
                    logger.error(e)

        return logprobs

    @retry(stop=stop_after_attempt(5), wait=wait_exponential(multiplier=1, min=2))
    def _generate(self, prompt: Prompt) -> Generator[LLMEvent, None, None]:
        self.load_tokenizer()
        headers = {"Content-Type": "application/json"}
        if self.api_token:
            headers |= {"Authorization": f"Bearer {self.api_token}"}
        data = {
            "model": self.model_name,
            "messages": prompt.messages,
            "stream": self.stream,
            "tools": prompt.tools,
        }
        if self.collect_logprobs:
            data.update(
                {
                    "logprobs": 1,
                    "include_stop_str_in_output": True,
                    "skip_special_tokens": False,
                }
            )
        logger.debug(f"POST request to {self.base_url}/v1/chat/completions")
        start_send_request = time.time()
        for k, v in self.parameters.items():
            data[k] = OmegaConf.to_container(v) if isinstance(v, DictConfig) else v
        r = requests.post(
            url=f"{self.base_url}/v1/chat/completions",
            json=data,
            headers=headers,
            stream=self.stream,
            verify=False,
        )
        time_send_request = time.time() - start_send_request
        self._stats["time_send_request"].append(time_send_request)
        if not r.ok:
            logger.error(f"Failed to get completion: {r.text}")
            r.raise_for_status()
        parsed_logprobs = []
        if self.stream:
            response_buffer = []
            for byte_payload in r.iter_lines():
                if byte_payload == b"\n":
                    continue
                payload = byte_payload.decode("utf-8")
                if payload.startswith("data:"):
                    if payload == "data: [DONE]":
                        continue
                    json_payload = json.loads(payload.lstrip("data:").rstrip("\n"))
                    response_delta = json_payload["choices"][0]["delta"].get("content", "")
                    if not response_delta:
                        continue
                    response_buffer.append(response_delta)
                    yield LLMEvent(chunk=response_delta)
            output = LLMOutput(content="".join(response_buffer))
        else:
            data = r.json()
            try:
                content = data["choices"][0]["message"]["content"]
                tool_calls = data["choices"][0]["message"].get("tool_calls", [])
                if not content and not tool_calls:
                    logger.warning(f"Empty completion {data}")

                if self.collect_logprobs:
                    completion_logprobs = data["choices"][0]["logprobs"]["content"]
                    parsed_logprobs = self.parse_completion_logprobs(completion_logprobs)
            except Exception as e:
                logger.exception(f"Failed to parse llm response: {r}")
                raise e
            output = LLMOutput(content=content)
            if tool_calls:
                output.tool_calls = [litellm.ChatCompletionMessageToolCall(**tc) for tc in tool_calls]
        llm_call = self.log_output(prompt, output)
        llm_call.logprobs = parsed_logprobs
        yield LLMEvent(output=output, llm_call=llm_call)

    def get_step_schema(self, cls):
        return get_step_schemas_from_union_type(cls)

    def load_tokenizer(self):
        """
        Loads the tokenizer for the model.

        If the tokenizer is not already loaded, this method will import the
        `transformers` library and load the tokenizer using the model name or
        tokenizer name. If `_MOCK_TOKENIZER` is set, it will use that instead.

        Raises:
            ValueError: If neither `self.tokenizer_name` nor `self.model_name`
                        is provided and `_MOCK_TOKENIZER` is not set.
        """
        if self.tokenizer is None:
            global transformers
            if transformers is None:
                import transformers
            name = _MOCK_TOKENIZER if _MOCK_TOKENIZER else (self.tokenizer_name or self.model_name)
            try:
                self.tokenizer = transformers.AutoTokenizer.from_pretrained(name)
            except Exception as e:
                if not self.use_litellm_tokenizer_fallback:
                    raise e

    def make_training_text(self, prompt: Prompt, output: LLMOutput) -> TrainingText:
        """
        Generates training text from a given prompt and LLM output.

        This method loads the tokenizer and uses it to create training text
        suitable for training a language model.

        Args:
            prompt (Prompt): The input prompt to generate training text from.
            output (LLMOutput): The output from the language model to be used in training.

        Returns:
            TrainingText: The generated training text.
        """
        self.load_tokenizer()
        return trainable_llm_make_training_text(prompt, output, self.tokenizer)

    def get_logprobs_token_ids(self, prompt_token_ids: list[int], completion_token_ids: list[int]) -> dict[str, Any]:
        if not self.tokenizer:
            self.load_tokenizer()

        headers = {"Content-Type": "application/json"}
        if self.api_token:
            headers |= {"Authorization": f"Bearer {self.api_token}"}

        generation_args = {
            "model": self.model_name,
            "prompt": prompt_token_ids + completion_token_ids,
            "temperature": 0.0,
            "max_tokens": 0,
            "logprobs": 0,
            "echo": True,
            "include_stop_str_in_output": True,  # self.include_stop_str_in_output,
            "skip_special_tokens": False,
            "n": 1,  # number of completions to generate
            "stream": False,  # return a single completion and not a stream of lines
        }
        url = f"{self.base_url}/v1/completions"
        logger.debug(f"POST request to {url}")
        r = requests.post(url, json=generation_args, headers=headers, verify=False)
        r.raise_for_status()  # raise exception if status code is not in the 200s
        try:
            response = r.json()
        except Exception as e:
            raise RuntimeError(f"Generation API wrong response: {r.text}", e)
        logprobs = []
        completion_logprobs = response["choices"][0]["prompt_logprobs"][-len(completion_token_ids) :]
        for lp in completion_logprobs:
            if lp:
                for k, v in lp.items():
                    v.update({"generated": 0, "token_id": k})
                    logprobs.append(v)
        return {"content": logprobs}

    def get_batch_logprobs_token_ids(
        self, prompt_token_ids: list[list[int]], completion_token_ids: list[list[int]]
    ) -> list[dict[str, Any]]:
        if not self.tokenizer:
            self.load_tokenizer()
        batch_size = len(prompt_token_ids)

        headers = {"Content-Type": "application/json"}
        if self.api_token:
            headers |= {"Authorization": f"Bearer {self.api_token}"}

        generation_args = {
            "model": self.model_name,
            "prompt": [pids + cids for pids, cids in zip(prompt_token_ids, completion_token_ids)],
            "temperature": 0.0,
            "max_tokens": 0,
            "logprobs": 0,
            "echo": True,
            "include_stop_str_in_output": True,  # self.include_stop_str_in_output,
            "skip_special_tokens": False,
            "n": 1,  # number of completions to generate
            "stream": False,  # return a single completion and not a stream of lines
        }
        url = f"{self.base_url}/v1/completions"
        logger.debug(f"POST request to {url}")
        r = requests.post(url, json=generation_args, headers=headers, verify=False)
        r.raise_for_status()

        try:
            response = r.json()
        except Exception as e:
            raise RuntimeError(f"Generation API wrong response: {r.text}", e)

        all_logprobs = []
        for i in range(batch_size):
            logprobs = []
            for lp in response["choices"][i]["prompt_logprobs"][-len(completion_token_ids[i]) :]:
                if lp:
                    for k, v in lp.items():
                        v.update({"generated": 0, "token_id": k})
                        logprobs.append(v)
            all_logprobs.append({"content": logprobs})
        return all_logprobs

    def get_logprobs_complete(self, prompt: str, output: str) -> dict[str, Any]:
        """
        Get the log probabilities of the tokens in the output given the prompt.

        This method sends a request to the language model API to generate the log probabilities
        for the tokens in the provided output, given the prompt. It uses the tokenizer to encode
        the prompt and output, and extracts the log probabilities from the API response.

        Args:
            prompt (str): The input prompt text.
            output (str): The output text for which log probabilities are to be calculated.

        Returns:
            list[float]: A list of log probabilities for each token in the output.

        Raises:
            RuntimeError: If the API response is not as expected or if there is a mismatch
                          between the tokens in the response and the provided output.
        """
        if not self.tokenizer:
            self.load_tokenizer()

        headers = {"Content-Type": "application/json"}
        if self.api_token:
            headers |= {"Authorization": f"Bearer {self.api_token}"}

        if self.tokenizer.bos_token and prompt.startswith(self.tokenizer.bos_token):
            prompt = prompt[len(self.tokenizer.bos_token) :]

        prompt_text = prompt + output
        generation_args = {
            "model": self.model_name,
            "prompt": prompt_text,
            "temperature": 0.0,
            "max_tokens": 0,
            "logprobs": 1,
            "echo": True,
            "include_stop_str_in_output": True,  # self.include_stop_str_in_output,
            "skip_special_tokens": False,
            "n": 1,  # number of completions to generate
            "stream": False,  # return a single completion and not a stream of lines
        }
        base_url = self.get_base_url()
        url = f"{base_url}/v1/completions"
        logger.debug(f"POST request to {url}")
        r = requests.post(url, json=generation_args, headers=headers, verify=False)
        r.raise_for_status()  # raise exception if status code is not in the 200s
        try:
            response = r.json()
            tokens = response["choices"][0]["logprobs"]["tokens"]
            log_probs = response["choices"][0]["logprobs"]["token_logprobs"]
            prompt_encoded = self.tokenizer.encode(prompt, add_special_tokens=True)
            prompt_completion_encoded = self.tokenizer.encode(prompt + output, add_special_tokens=True)
            completion_log_probs = log_probs[len(prompt_encoded) : len(prompt_completion_encoded)]
            completion_tokens = tokens[len(prompt_encoded) : len(prompt_completion_encoded)]
            assert (
                "".join(completion_tokens) == output
            ), f"Tokens do not match completion: {''.join(completion_tokens)} != {output}"
        except Exception as e:
            raise RuntimeError(f"Generation API wrong response: {r.text}", e)
        completion_log_probs = [
            {
                "logprob": lp,
                "top_logprobs": [],
                "token": t,
            }
            for lp, t in zip(completion_log_probs, completion_tokens)
        ]
        return {"content": completion_log_probs}

    def get_logprobs_chat_complete(self, prompt: Prompt, output: LLMOutput) -> dict[str, Any]:
        """
        Calculate the log probabilities of the tokens in the completion generated by the language model.

        This function sends a request to the language model API to generate completions and calculate log probabilities.
        The function uses the tokenizer to encode the prompt and completion texts.
        The log probabilities are extracted from the API response and validated against the original completion.

        Args:
            prompt (Prompt): The prompt containing the messages to be sent to the language model.
            output (LLMOutput): The output from the language model containing the generated completion.

        Returns:
            list[float]: A list of log probabilities for each token in the generated completion.

        Raises:
            RuntimeError: If the response from the generation API is incorrect or cannot be parsed.
        """
        headers = {"Content-Type": "application/json"}
        if self.api_token:
            headers |= {"Authorization": f"Bearer {self.api_token}"}

        time_t0 = time.time()
        prompt_text = self.tokenizer.apply_chat_template(prompt.messages, tokenize=False)
        completion = output.content or ""
        messages = prompt.messages + [{"role": "assistant", "content": completion}]
        prompt_text = self.tokenizer.apply_chat_template(prompt.messages, tokenize=False, add_generation_prompt=True)
        prompt_completion_text = self.tokenizer.apply_chat_template(messages, tokenize=False)
        if self.tokenizer.bos_token and prompt_text.startswith(self.tokenizer.bos_token):
            prompt_text = prompt_text[len(self.tokenizer.bos_token) :]
            prompt_completion_text = prompt_completion_text[len(self.tokenizer.bos_token) :]

        prompt_encoded = self.tokenizer.encode(prompt_text, add_special_tokens=False)
        prompt_completion_encoded = self.tokenizer.encode(prompt_completion_text, add_special_tokens=False)

        generation_args = {
            "model": self.model_name,
            "messages": messages,
            "temperature": 0.0,
            "max_tokens": 1,
            "logprobs": 1,
            "echo": True,
            "include_stop_str_in_output": True,  # self.include_stop_str_in_output,
            "skip_special_tokens": False,
            "n": 1,  # number of completions to generate
            "stream": False,  # return a single completion and not a stream of lines
        }
        base_url = self.get_base_url()
        r = requests.post(
            url=f"{base_url}/v1/chat/completions",
            json=generation_args,
            headers=headers,
            verify=False,
        )
        r.raise_for_status()

        try:
            response = r.json()
            log_probs = [list(log_prob.values())[0] for log_prob in response["prompt_logprobs"] if log_prob]
            completion_log_probs = log_probs[len(prompt_encoded) : len(prompt_completion_encoded)]
            decoded_completion_tokens = [log_prob["decoded_token"] for log_prob in completion_log_probs]
            reconstructed_completion = "".join(decoded_completion_tokens)
            if self.tokenizer.eos_token in reconstructed_completion:
                reconstructed_completion = reconstructed_completion[: -len(self.tokenizer.eos_token)]
            assert (
                reconstructed_completion == completion
            ), f"Tokens do not match completion: {reconstructed_completion} != {completion}"
        except Exception as e:
            raise RuntimeError(f"Generation API wrong response: {r.text}", e)

        logger.debug(f"Log likelihood calculation took {time.time() - time_t0:.2f} seconds")
        logger.debug(f"Tokens per second: {len(log_probs) / (time.time() - time_t0):.2f}")

        completion_log_probs = [
            {
                "logprob": o["logprob"],
                "top_logprobs": [],
                "token": o["decoded_token"],
            }
            for o in completion_log_probs
        ]

        return {"content": completion_log_probs}

    @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=2))
    def get_logprobs(self, prompt: str | Prompt | list[int], output: str | LLMOutput | list[int]) -> dict[str, Any]:
        """
        Calculate the log probabilities of the given output based on the provided prompt.

        Args:
            prompt (Union[str, Prompt]): The input prompt, which can be either a string or a Prompt object.
            output (Union[str, LLMOutput]): The output to evaluate, which can be either a string or an LLMOutput object.

        Returns:
            list[float]: A list of log probabilities corresponding to the given output.

        Raises:
            ValueError: If the input types are not valid.
        """
        if isinstance(prompt, str) and isinstance(output, str):
            return self.get_logprobs_complete(prompt=prompt, output=output)
        elif isinstance(prompt, Prompt) and isinstance(output, LLMOutput):
            return self.get_logprobs_chat_complete(prompt=prompt, output=output)
        elif isinstance(prompt, list) and isinstance(output, list):
            return self.get_logprobs_token_ids(prompt_token_ids=prompt, completion_token_ids=output)
        else:
            raise ValueError("Invalid input types")

    def count_tokens(self, messages: list[dict] | str) -> int:
        """
        Count the number of tokens in the given messages.

        This method loads the tokenizer and then counts the number of tokens
        in the provided messages. The messages can be either a string or a list
        of dictionaries.

        Args:
            messages (Union[list[dict], str]): The messages to count tokens for. It can
                               be a single string or a list of dictionaries.

        Returns:
            int: The number of tokens in the provided messages.
        """
        try:
            self.load_tokenizer()
            if isinstance(messages, str):
                return len(self.tokenizer(messages).input_ids)
            else:
                add_generation_prompt = False if messages[-1]["role"] == "assistant" else True
                return len(self.tokenizer.apply_chat_template(messages, add_generation_prompt=add_generation_prompt))
        except Exception as e:
            if self.use_litellm_tokenizer_fallback:
                logger.warning("Failed to count tokens with tokenizer, fallback to litellm counter")
                if isinstance(messages, str):
                    return litellm.token_counter(model=self.model_name, text=messages)  # type: ignore
                else:
                    return litellm.token_counter(model=self.model_name, messages=messages)  # type: ignore
            else:
                raise e

    async def agenerate(self, prompt: Prompt, session: aiohttp.ClientSession, **kwargs) -> LLMStream:
        headers = {"Content-Type": "application/json"}
        if self.api_token:
            headers |= {"Authorization": f"Bearer {self.api_token}"}
        params = {"model": self.model_name, "messages": prompt.messages, "tools": prompt.tools} | kwargs
        for k, v in self.parameters.items():
            params[k] = OmegaConf.to_container(v) if isinstance(v, DictConfig) else v
        if self.collect_logprobs:
            params.update(
                {
                    "logprobs": 1,
                    "include_stop_str_in_output": True,
                    "skip_special_tokens": False,
                }
            )

        logger.debug(
            f"POST request to {self.base_url}/v1/chat/completions with params: {pprint.pformat(params, width=120)}"
        )
        async with self._semaphore:  # type: ignore
            retry_count = 0
            while True:
                try:
                    async with session.post(
                        url=f"{self.base_url}/v1/chat/completions", json=params, headers=headers, ssl=False
                    ) as response:
                        if not response.ok:
                            error_text = await response.text()
                            logger.error(f"Failed to get completion: {error_text}")
                            response.raise_for_status()
                        data = await response.json()
                        break
                except asyncio.TimeoutError as e:
                    logger.exception("API Timeout, retrying in 1 sec")
                    retry_count += 1
                    if retry_count > self.max_retries:
                        raise e
                    delay = self.base_delay * (2 ** (retry_count - 1))
                    logger.warning(
                        f"API Timeout, retrying in {delay:.2f} seconds (attempt {retry_count}/{self.max_retries})"
                    )
                    await asyncio.sleep(delay)
                except aiohttp.ClientError as e:
                    logger.error(f"Connection error for {self.base_url}/v1/chat/completions: {e}")
                    retry_count += 1
                    if retry_count > self.max_retries:
                        raise e
                    delay = self.base_delay * (2 ** (retry_count - 1))
                    logger.warning(f"Retrying in {delay:.2f} seconds (attempt {retry_count}/{self.max_retries})")
                    await asyncio.sleep(delay)

        try:
            content = data["choices"][0]["message"]["content"]
            tool_calls = data["choices"][0]["message"].get("tool_calls", [])
            if not content and not tool_calls:
                logger.warning(f"Empty completion {data}")

            parsed_logprobs = []
            if self.collect_logprobs:
                completion_logprobs = data["choices"][0]["logprobs"]["content"]
                for logprob in completion_logprobs:
                    if logprob:
                        try:
                            # We assume that the server was launched with --return-tokens-as-token-ids
                            # and that the tokens are provided as: ['token_id:1271', 'token_id:1505', '
                            parsed_logprobs.append(
                                TokenLogprob(
                                    token_id=int(logprob["token"].split(":")[-1]),
                                    logprob=logprob["logprob"],
                                )
                            )
                        except Exception as e:
                            logger.error(f"Failed to process logprobs: {logprob}")
                            logger.error(e)
        except Exception as e:
            logger.exception(f"Failed to parse llm response: {data}")
            raise e

        prompt_tokens = data["usage"]["prompt_tokens"]
        completion_tokens = data["usage"]["completion_tokens"]

        output = LLMOutput(content=content or "")
        if tool_calls:
            output.tool_calls = [litellm.ChatCompletionMessageToolCall(**tc) for tc in tool_calls]
        llm_call = self.log_output(
            prompt,
            output,
            prompt_length_tokens=prompt_tokens,
            output_length_tokens=completion_tokens,
            count_tokens=False,
        )
        assert llm_call is not None, "llm_call is None"
        llm_call.logprobs = parsed_logprobs

        def _gen():
            yield LLMEvent(llm_call=llm_call, output=output)

        return LLMStream(generator=_gen(), prompt=prompt)



import copy

import pytest
from pydantic import SerializeAsAny

from tapeagents.agent import DEFAULT, Agent, AgentEvent, AgentStream, Node
from tapeagents.core import Action, AgentStep, PartialStep, Prompt, StepMetadata, Tape, Thought
from tapeagents.llms import LLMStream, MockLLM

MockTape = Tape[None, Action]
MockAgentEvent = AgentEvent[MockTape]


class MockAgent(Agent):
    pass


class MockPrompt(Prompt):
    pass


class EmptyLLM(MockLLM):
    def generate(self, prompt):
        return LLMStream(None, prompt)


def test_get_final_tape_success():
    final_tape = MockTape(steps=[Action()])

    events = iter([MockAgentEvent(), MockAgentEvent(final_tape=final_tape)])
    agent_stream = AgentStream[MockTape](events)  # type: ignore

    assert agent_stream.get_final_tape() == final_tape


def test_get_final_tape_failure():
    events = iter([MockAgentEvent(), MockAgentEvent()])
    agent_stream = AgentStream(events)  # type: ignore

    with pytest.raises(ValueError, match="Agent didn't produce final tape"):
        agent_stream.get_final_tape()


def test_node_make_prompt():
    class MockNode(Node):
        def make_prompt(self, agent, tape):
            return MockPrompt()

    node = MockNode()
    agent = MockAgent()
    tape = MockTape()

    prompt = node.make_prompt(agent, tape)

    assert isinstance(prompt, MockPrompt)


def test_node_generate_steps():
    class MockNode(Node):
        def generate_steps(self, agent, tape, llm_stream):
            yield PartialStep(step=Action())
            yield Action()

    node = MockNode()
    agent = MockAgent()
    tape = MockTape()
    llm_stream = LLMStream(None, Prompt())

    steps = list(node.generate_steps(agent, tape, llm_stream))

    assert len(steps) == 2
    assert isinstance(steps[0], PartialStep)
    assert isinstance(steps[1], Action)


def test_model_post_init_no_name():
    agent = MockAgent()
    agent.model_post_init(None)
    assert agent.name == "MockAgent"


def test_model_post_init_duplicate_subagent_name():
    subagent1 = MockAgent(name="subagent")
    subagent2 = MockAgent(name="subagent")

    with pytest.raises(ValueError, match='Duplicate subagent name "subagent" in subagent 1'):
        MockAgent(subagents=[subagent1, subagent2])


def test_model_post_init_subagent_already_has_manager():
    subagent = MockAgent(name="subagent")
    subagent._manager = MockAgent(name="manager")

    with pytest.raises(ValueError, match="Agent is already a subagent of another agent. Make a copy of your agent."):
        MockAgent(subagents=[subagent])


def test_model_post_init_subagent_not_instance_of_agent():
    subagent = "not_an_agent"

    with pytest.raises(ValueError, match="Subagents must be instances of Agent"):
        MockAgent(subagents=[subagent])


def test_model_post_init_duplicate_node_name():
    node1 = Node(name="node")
    node2 = Node(name="node")

    with pytest.raises(ValueError, match='Duplicate node name "node" in node 1'):
        Agent(nodes=[node1, node2])


def test_create_with_llm():
    llm = MockLLM()
    agent = MockAgent.create(llm)

    assert isinstance(agent, MockAgent)
    assert agent.llms[DEFAULT] == llm


def test_create_with_templates():
    template = "template_string"
    agent = MockAgent.create(templates=template)

    assert isinstance(agent, MockAgent)
    assert agent.templates[DEFAULT] == template


def test_create_with_llms_and_templates():
    llm = MockLLM()
    template = "template_string"
    agent = MockAgent.create(llm, templates=template)

    assert isinstance(agent, MockAgent)
    assert agent.llms[DEFAULT] == llm
    assert agent.templates[DEFAULT] == template


def test_create_with_kwargs():
    agent = MockAgent.create(name="test_agent", max_iterations=10)

    assert isinstance(agent, MockAgent)
    assert agent.name == "test_agent"
    assert agent.max_iterations == 10


def test_create_with_dict_llms_and_templates():
    llms = {"default": MockLLM(), "llm2": MockLLM()}
    templates = {"template1": "template_string1", "template2": "template_string2"}
    agent = MockAgent.create(llms, templates=templates)  # type: ignore

    assert isinstance(agent, MockAgent)
    assert agent.llms == llms
    assert agent.templates == templates


def test_make_prompt():
    class NodeMockPrompt(Prompt):
        pass

    class MockNode(Node):
        def make_prompt(self, agent, tape):
            return NodeMockPrompt()

    class MockAgent(Agent):
        def select_node(self, tape):
            return MockNode()

    agent = MockAgent()
    tape = MockTape()

    prompt = agent.make_prompt(tape)

    assert isinstance(prompt, NodeMockPrompt)


def test_generate_steps():
    class MockNode(Node):
        def generate_steps(self, agent, tape, llm_stream):
            yield PartialStep(step=Action())
            yield Action()

    class MockAgent(Agent):
        def select_node(self, tape):
            return MockNode()

    agent = MockAgent()
    tape = MockTape()
    llm_stream = LLMStream(None, Prompt())

    steps = list(agent.generate_steps(tape, llm_stream))

    assert len(steps) == 2
    assert isinstance(steps[0], PartialStep)
    assert isinstance(steps[1], Action)


def test_run_iteration_with_llm_stream():
    class MockNode(Node):
        def generate_steps(self, agent, tape, llm_stream):
            yield PartialStep(step=Action())
            yield Action()

    class MockAgent(Agent):
        def select_node(self, tape):
            return MockNode()

    agent = MockAgent()
    tape = MockTape()
    llm_stream = LLMStream(None, Prompt())

    steps = list(agent.run_iteration(tape, llm_stream))

    assert len(steps) == 2
    assert isinstance(steps[0], PartialStep)
    assert isinstance(steps[1], Action)


def test_run_iteration():
    class MockNode(Node):
        def generate_steps(self, agent, tape, llm_stream):
            yield PartialStep(step=Action())
            yield Action()

    class MockAgent(Agent):
        def select_node(self, tape):
            return MockNode()

    agent = MockAgent(llms={DEFAULT: EmptyLLM()})
    tape = MockTape()

    steps = list(agent.run_iteration(tape))

    assert len(steps) == 2
    assert isinstance(steps[0], PartialStep)
    assert isinstance(steps[1], Action)


def test_run_iteration_with_multiple_llms():
    class MockNode(Node):
        def generate_steps(self, agent, tape, llm_stream):
            yield PartialStep(step=Action())
            yield Action()

    class MockAgent(Agent):
        def select_node(self, tape):
            return MockNode()

    agent = MockAgent(llms={"default": EmptyLLM(), "llm2": EmptyLLM()})
    tape = MockTape()

    for step in agent.run_iteration(tape):
        assert isinstance(step, (PartialStep, Action))
        if not isinstance(step, PartialStep):
            assert step.metadata.llm == "default"


def test_run_iteration_with_agent_step():
    class MockNode(Node):
        def generate_steps(self, agent, tape, llm_stream):
            yield AgentStep(metadata=StepMetadata())

    class MockAgent(Agent):
        nodes: list[SerializeAsAny[Node]] = [MockNode(name="node1")]

        def select_node(self, tape):
            return self.nodes[0]

        def get_node_name(self, tape) -> str:
            return self.nodes[0].name

    agent = MockAgent(llms={DEFAULT: EmptyLLM()})
    tape = MockTape()

    steps = list(agent.run_iteration(tape))

    assert len(steps) == 1
    assert isinstance(steps[0], AgentStep)
    assert steps[0].metadata.prompt_id is not None
    assert steps[0].metadata.node == "node1"


def test_select_node_explicit_next_node():
    class MockView:
        next_node = "node2"
        last_node = None

    class MockTapeViewStack:
        top = MockView()

    class MockAgent(Agent):
        def compute_view(self, tape):
            return MockTapeViewStack()

        def find_node(self, name):
            return Node(name=name)

    agent = MockAgent(nodes=[Node(name="node1"), Node(name="node2")])
    tape = MockTape()

    selected_node = agent.select_node(tape)

    assert selected_node.name == "node2"


def test_select_node_no_nodes_run_yet():
    class MockView:
        next_node = None
        last_node = None

    class MockTapeViewStack:
        top = MockView()

    class MockAgent(Agent):
        def compute_view(self, tape):
            return MockTapeViewStack()

    agent = MockAgent(nodes=[Node(name="node1"), Node(name="node2")])
    tape = MockTape()

    selected_node = agent.select_node(tape)

    assert selected_node.name == "node1"


def test_select_node_select_next_node():
    class MockView:
        next_node = None
        last_node = "node1"

    class MockTapeViewStack:
        top = MockView()

    class MockAgent(Agent):
        def compute_view(self, tape):
            return MockTapeViewStack()

    agent = MockAgent(nodes=[Node(name="node1"), Node(name="node2")])
    tape = MockTape()

    selected_node = agent.select_node(tape)

    assert selected_node.name == "node2"


def test_select_node_next_node_not_found():
    class MockView:
        next_node = None
        last_node = "node2"

    class MockTapeViewStack:
        top = MockView()

    class MockAgent(Agent):
        def compute_view(self, tape):
            return MockTapeViewStack()

    agent = MockAgent(nodes=[Node(name="node1"), Node(name="node2")])
    tape = MockTape()

    with pytest.raises(ValueError, match="Next node not found"):
        agent.select_node(tape)


def test_run_metadata():
    class MockNode(Node):
        def generate_steps(self, agent, tape, llm_stream):
            yield Thought()
            yield Action()

    class MockAgent(Agent):
        def select_node(self, tape):
            return MockNode()

    agent = MockAgent(llms={DEFAULT: EmptyLLM()})
    tape = MockTape()

    initial_tape_metadata = copy.deepcopy(tape.metadata)
    assert initial_tape_metadata.n_added_steps == 0

    final_tape = agent.run(tape).get_final_tape()

    # check that the original tape metadata is the same
    assert tape.metadata == initial_tape_metadata
    # check that the new tape metadata is updated correclty
    assert final_tape.metadata.id != initial_tape_metadata.id
    assert final_tape.metadata.parent_id == initial_tape_metadata.id
    assert final_tape.metadata.n_added_steps == 2
    assert final_tape.metadata.author == agent.name
    # assert that the rest is the same, except for id, parent_id, n_added_steps, and author
    initial_tape_metadata.id, final_tape.metadata.id = None, None
    initial_tape_metadata.parent_id, final_tape.metadata.parent_id = None, None
    initial_tape_metadata.n_added_steps, final_tape.metadata.n_added_steps = 0, 0
    initial_tape_metadata.author, final_tape.metadata.author = None, None
    assert initial_tape_metadata == final_tape.metadata



import re
import string
import warnings


def normalize_number_str(number_str: str) -> float:
    # we replace these common units and commas to allow
    # conversion to float
    for char in ["$", "%", ","]:
        number_str = number_str.replace(char, "")
    try:
        return float(number_str)
    except ValueError:
        print(f"String {number_str} cannot be normalized to number str.")
        return float("inf")


def split_string(
    s: str,
    char_list: list[str] = [",", ";"],
) -> list[str]:
    pattern = f"[{''.join(char_list)}]"
    return re.split(pattern, s)


def question_scorer(
    model_answer: str,
    ground_truth: str,
) -> bool:
    def is_float(element: any) -> bool:
        try:
            float(element)
            return True
        except ValueError:
            return False

    # if gt is a number
    if is_float(ground_truth):
        print(f"Evaluating {model_answer} as a number.")
        normalized_answer = normalize_number_str(model_answer)
        return normalized_answer == float(ground_truth)

    # if gt is a list
    elif any(char in ground_truth for char in [",", ";"]):
        print(f"Evaluating {model_answer} as a comma separated list.")
        # question with the fish: normalization removes punct

        gt_elems = split_string(ground_truth)
        ma_elems = split_string(model_answer)

        # check length is the same
        if len(gt_elems) != len(ma_elems):
            warnings.warn("Answer lists have different lengths, returning False.", UserWarning)
            return False

        # compare each element as float or str
        comparisons = []
        for ma_elem, gt_elem in zip(ma_elems, gt_elems):
            if is_float(gt_elem):
                normalized_ma_elem = normalize_number_str(ma_elem)
                comparisons.append(normalized_ma_elem == float(gt_elem))
            else:
                # we do not remove punct since comparisons can include punct
                comparisons.append(
                    normalize_str(ma_elem, remove_punct=False) == normalize_str(gt_elem, remove_punct=False)
                )
        return all(comparisons)

    # if gt is a str
    else:
        print(f"Evaluating {model_answer} as a string.")
        return normalize_str(model_answer) == normalize_str(ground_truth)


def normalize_str(input_str, remove_punct=True) -> str:
    """
    Normalize a string by:
    - Removing all white spaces
    - Optionally removing punctuation (if remove_punct is True)
    - Converting to lowercase
    Parameters:
    - input_str: str, the string to normalize
    - remove_punct: bool, whether to remove punctuation (default: True)
    Returns:
    - str, the normalized string
    """
    # Remove all white spaces. Required e.g for seagull vs. sea gull
    no_spaces = re.sub(r"\s", "", input_str)

    # Remove punctuation, if specified.
    if remove_punct:
        translator = str.maketrans("", "", string.punctuation)
        return no_spaces.lower().translate(translator)
    else:
        return no_spaces.lower()



import pathlib
import shutil
from pathlib import Path

import testbook

from tests.make_test_data import run_test_in_tmp_dir

res_dir = f"{pathlib.Path(__file__).parent.resolve()}/res"


def test_intro_notebook():
    intro_notebook_path = Path("intro.ipynb").resolve()
    assets_path = Path("assets").resolve()
    with testbook.testbook(intro_notebook_path) as tb:
        with run_test_in_tmp_dir("tests/examples/res/intro_notebook") as test_data_dir:
            shutil.copytree(assets_path, Path("assets"))
            sqlite_path = Path(test_data_dir) / "tapedata.sqlite"
            cache_dir = Path(test_data_dir) / "cache"
            tb.inject(
                f"""
                import os
                from tapeagents import llms
                os.environ["TAPEAGENTS_SQLITE_DB"] = "{sqlite_path}"
                os.environ["TAPEAGENTS_CACHE_DIR"] = "{cache_dir}"
                os.environ["TAPEAGENTS_FORCE_CACHE"] = "1"
                llms.cached._REPLAY_SQLITE = "{res_dir}/intro_notebook/tapedata.sqlite"
                """,
                before=0,
            )
            tb.execute()


if __name__ == "__main__":
    test_intro_notebook()



# Modified from the original source: https://github.com/microsoft/autogen/blob/gaia_multiagent_v01_march_1st/autogen/mdconvert.py
# MIT License

# Copyright (c) Microsoft Corporation.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import base64
import copy
import html
import json
import logging
import mimetypes
import os
import re
import shutil
import subprocess
import sys
import tempfile
import traceback
from typing import Any, List, Optional, Union
from urllib.parse import parse_qs, urlparse

import markdownify
import numpy as np
import pdfminer
import pdfminer.high_level
import PIL
import pptx
import puremagic
import requests
import whisper
from bs4 import BeautifulSoup
from docling.datamodel.base_models import InputFormat
from docling.datamodel.pipeline_options import PdfPipelineOptions, TableFormerMode, TableStructureOptions
from docling.document_converter import DocumentConverter as DoclingDocumentConverter, PdfFormatOption
from pydantic import BaseModel, Field
from readability import Document
from youtube_transcript_api import YouTubeTranscriptApi

logger = logging.getLogger(__name__)


class DocumentConverterResult:
    """The result of converting a document to text."""

    def __init__(self, title: Union[str, None] = None, text_content: str = ""):
        self.title = title
        self.text_content = text_content


class DocumentConverter(BaseModel):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        raise NotImplementedError()


class DoclingConverter(DocumentConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        converter = DoclingDocumentConverter(
            allowed_formats=kwargs.get("allowed_formats", None), format_options=kwargs.get("format_options", None)
        )
        result = converter.convert(local_path, page_range=(1, kwargs.get("max_pages", 20)))
        markdown = result.document.export_to_markdown()
        return DocumentConverterResult(
            title=None,
            text_content=markdown,
        )


class PlainTextConverter(DocumentConverter):
    """Anything with content type text/plain"""

    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        extension = kwargs.get("file_extension", "")
        if extension == "":
            return None

        content_type, encoding = mimetypes.guess_type("__placeholder" + extension)
        if content_type is None:
            return None

        if "text/" not in content_type.lower() and extension.lower() not in [".txt", ".xml", ".jsonld", ".pdb"]:
            return None

        text_content = ""
        with open(local_path, "rt") as fh:
            text_content = fh.read()

        return DocumentConverterResult(
            title=None,
            text_content=text_content,
        )


class HtmlConverter(DocumentConverter):
    """Anything with content type text/html"""

    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not html
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".html", ".htm"]:
            return None

        result = None
        readability = kwargs.get("readability", False)
        strip_links = kwargs.get("strip_links", False)
        strip_images = kwargs.get("strip_images", False)
        with open(local_path, "rt") as fh:
            result = self._convert(fh.read(), readability, strip_links, strip_images)

        return result

    def _convert(
        self, html_content, readability: bool = False, strip_links: bool = False, strip_images: bool = False
    ) -> Union[None, DocumentConverterResult]:
        """Helper function that converts and HTML string."""

        # Parse the string
        soup = BeautifulSoup(html_content, "html.parser")

        # Remove javascript and style blocks
        for script in soup(["script", "style"]):
            script.extract()

        # Print only the main content
        body_elm = soup.find("body")
        strip = []
        if strip_links:
            strip.append("a")
        if strip_images:
            strip.append("img")
        webpage_text = ""
        if body_elm:
            webpage_text = markdownify.MarkdownConverter(strip=strip).convert_soup(body_elm)
        else:
            webpage_text = markdownify.MarkdownConverter(strip=strip).convert_soup(soup)
        if readability and len(webpage_text) > 4000:
            doc = Document(soup.prettify())
            clean_html = doc.summary()
            webpage_text = markdownify.markdownify(clean_html, strip=strip)

        return DocumentConverterResult(
            title=None if soup.title is None else soup.title.string, text_content=webpage_text
        )


class HtmlDoclingConverter(DoclingConverter):
    """Anything with content type text/html"""

    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not html
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".html", ".htm"]:
            return None

        return super().convert(local_path, **kwargs)


class WikipediaConverter(DocumentConverter):
    """Handle Wikipedia pages separately, focusing only on the main document content."""

    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not Wikipedia
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".html", ".htm"]:
            return None
        url = kwargs.get("url", "")
        if not re.search(r"^https?:\/\/[a-zA-Z]{2,3}\.wikipedia.org\/", url):
            return None

        # Parse the file
        soup = None
        with open(local_path, "rt") as fh:
            soup = BeautifulSoup(fh.read(), "html.parser")

        # Remove javascript and style blocks
        for script in soup(["script", "style"]):
            script.extract()

        # Print only the main content
        body_elm = soup.find("div", {"id": "mw-content-text"})
        title_elm = soup.find("span", {"class": "mw-page-title-main"})

        webpage_text = ""
        if body_elm:
            # What's the title
            main_title = soup.title.string
            if title_elm and len(title_elm) > 0:
                main_title = title_elm.string

            # Convert the page
            webpage_text = "# " + main_title + "\n\n" + markdownify.MarkdownConverter().convert_soup(body_elm)
        else:
            webpage_text = markdownify.MarkdownConverter().convert_soup(soup)

        return DocumentConverterResult(
            title=soup.title.string,
            text_content=webpage_text,
        )


class YouTubeConverter(DocumentConverter):
    """Handle YouTube specially, focusing on the video title, description, and transcript."""

    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not YouTube
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".html", ".htm"]:
            return None
        url = kwargs.get("url", "")
        if not url.startswith("https://www.youtube.com/watch?"):
            return None

        # Parse the file
        soup = None
        with open(local_path, "rt") as fh:
            soup = BeautifulSoup(fh.read(), "html.parser")

        # Read the meta tags
        metadata = {"title": soup.title.string}
        for meta in soup(["meta"]):
            for a in meta.attrs:
                if a in ["itemprop", "property", "name"]:
                    metadata[meta[a]] = meta.get("content", "")
                    break

        # We can also try to read the full description. This is more prone to breaking, since it reaches into the page implementation
        try:
            for script in soup(["script"]):
                content = script.text
                if "ytInitialData" in content:
                    lines = re.split(r"\r?\n", content)
                    obj_start = lines[0].find("{")
                    obj_end = lines[0].rfind("}")
                    if obj_start >= 0 and obj_end >= 0:
                        data = json.loads(lines[0][obj_start : obj_end + 1])
                        attrdesc = self._findKey(data, "attributedDescriptionBodyText")
                        if attrdesc:
                            metadata["description"] = attrdesc["content"]
                    break
        except Exception:
            pass

        # Start preparing the page
        webpage_text = "# YouTube\n"

        title = self._get(metadata, ["title", "og:title", "name"])
        if title:
            webpage_text += f"\n## {title}\n"

        stats = ""
        views = self._get(metadata, ["interactionCount"])
        if views:
            stats += f"- **Views:** {views}\n"

        keywords = self._get(metadata, ["keywords"])
        if keywords:
            stats += f"- **Keywords:** {keywords}\n"

        runtime = self._get(metadata, ["duration"])
        if runtime:
            stats += f"- **Runtime:** {runtime}\n"

        if len(stats) > 0:
            webpage_text += f"\n### Video Metadata\n{stats}\n"

        description = self._get(metadata, ["description", "og:description"])
        if description:
            webpage_text += f"\n### Description\n{description}\n"

        transcript_text = ""
        parsed_url = urlparse(url)
        params = parse_qs(parsed_url.query)
        if "v" in params:
            video_id = params["v"][0]
            try:
                # Must be a single transcript.
                transcript = YouTubeTranscriptApi.get_transcript(video_id)
                transcript_text = " ".join([part["text"] for part in transcript])
                # Alternative formatting:
                # formatter = TextFormatter()
                # formatter.format_transcript(transcript)
            except Exception:
                pass
        if transcript_text:
            webpage_text += f"\n### Transcript\n{transcript_text}\n"

        return DocumentConverterResult(
            title=title if title else soup.title.string,
            text_content=webpage_text,
        )

    def _get(self, json, keys, default=None):
        for k in keys:
            if k in json:
                return json[k]
        return default

    def _findKey(self, json, key):
        if isinstance(json, list):
            for elm in json:
                ret = self._findKey(elm, key)
                if ret is not None:
                    return ret
        elif isinstance(json, dict):
            for k in json:
                if k == key:
                    return json[k]
                else:
                    ret = self._findKey(json[k], key)
                    if ret is not None:
                        return ret
        return None


class PdfMinerConverter(DocumentConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a PDF
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".pdf":
            return None

        return DocumentConverterResult(
            title=None,
            text_content=pdfminer.high_level.extract_text(local_path),
        )


class PdfDoclingConverter(DoclingConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a PDF
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".pdf":
            return None
        pipeline_options = PdfPipelineOptions(
            do_table_structure=True,
            table_structure_options=TableStructureOptions(mode=TableFormerMode.ACCURATE),
        )
        format_options = {InputFormat.PDF: PdfFormatOption(pipeline_options=pipeline_options)}
        return super().convert(local_path, format_options=format_options, **kwargs)


class DocxConverter(HtmlConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a DOCX
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".docx":
            return None

        result = None
        import mammoth

        with open(local_path, "rb") as docx_file:
            result = mammoth.convert_to_html(docx_file)
            html_content = result.value
            result = self._convert(html_content)

        return result


class DocxDoclingConverter(DoclingConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a DOCX
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".docx":
            return None

        return super().convert(local_path, **kwargs)


class XlsxConverter(HtmlConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a XLSX
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".xlsx", ".xls"]:
            return None
        import pandas as pd

        sheets = pd.read_excel(local_path, sheet_name=None)
        md_content = ""
        for s in sheets:
            md_content += f"## {s}\n"
            html_content = sheets[s].to_html(index=False)
            md_content += self._convert(html_content).text_content.strip() + "\n\n"

        return DocumentConverterResult(
            title=None,
            text_content=md_content.strip(),
        )


class XlsxDoclingConverter(DoclingConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a XLSX
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".xlsx", ".xls"]:
            return None

        return super().convert(local_path, **kwargs)


class PptxConverter(HtmlConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a PPTX
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".pptx":
            return None

        md_content = ""

        presentation = pptx.Presentation(local_path)
        slide_num = 0
        for slide in presentation.slides:
            slide_num += 1

            md_content += f"\n\n<!-- Slide number: {slide_num} -->\n"

            title = slide.shapes.title
            for shape in slide.shapes:
                # Pictures
                if self._is_picture(shape):
                    # https://github.com/scanny/python-pptx/pull/512#issuecomment-1713100069
                    alt_text = ""
                    try:
                        alt_text = shape._element._nvXxPr.cNvPr.attrib.get("descr", "")
                    except Exception:
                        pass

                    # A placeholder name
                    filename = re.sub(r"\W", "", shape.name) + ".jpg"
                    # try:
                    #    filename = shape.image.filename
                    # except Exception:
                    #    pass

                    md_content += "\n![" + (alt_text if alt_text else shape.name) + "](" + filename + ")\n"

                # Tables
                if self._is_table(shape):
                    html_table = "<html><body><table>"
                    first_row = True
                    for row in shape.table.rows:
                        html_table += "<tr>"
                        for cell in row.cells:
                            if first_row:
                                html_table += "<th>" + html.escape(cell.text) + "</th>"
                            else:
                                html_table += "<td>" + html.escape(cell.text) + "</td>"
                        html_table += "</tr>"
                        first_row = False
                    html_table += "</table></body></html>"
                    md_content += "\n" + self._convert(html_table).text_content.strip() + "\n"

                # Text areas
                elif shape.has_text_frame:
                    if shape == title:
                        md_content += "# " + shape.text.lstrip() + " "
                    else:
                        md_content += shape.text + " "

            md_content = md_content.strip()

            if slide.has_notes_slide:
                md_content += "\n\n### Notes:\n"
                notes_frame = slide.notes_slide.notes_text_frame
                if notes_frame is not None:
                    md_content += notes_frame.text
                md_content = md_content.strip()

        return DocumentConverterResult(
            title=None,
            text_content=md_content.strip(),
        )

    def _is_picture(self, shape):
        if shape.shape_type == pptx.enum.shapes.MSO_SHAPE_TYPE.PICTURE:
            return True
        if shape.shape_type == pptx.enum.shapes.MSO_SHAPE_TYPE.PLACEHOLDER:
            if hasattr(shape, "image"):
                return True
        return False

    def _is_table(self, shape):
        if shape.shape_type == pptx.enum.shapes.MSO_SHAPE_TYPE.TABLE:
            return True
        return False


class PptxDoclingConverter(DoclingConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a PPTX
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".pptx":
            return None

        return super().convert(local_path, **kwargs)


class WavConverter(DocumentConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a WAV
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".wav":
            return None

        model = whisper.load_model("turbo")
        text_content = model.transcribe(local_path) or "[No speech detected]"
        return DocumentConverterResult(
            title=None,
            text_content=f"### Audio Transcript:\n{text_content}",
        )


class Mp3Converter(WavConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a MP3
        extension = kwargs.get("file_extension", "")
        if extension.lower() != ".mp3":
            return None
        import pydub

        handle, temp_path = tempfile.mkstemp(suffix=".wav")
        os.close(handle)
        try:
            sound = pydub.AudioSegment.from_mp3(local_path)
            sound.export(temp_path, format="wav")

            _args = dict()
            _args.update(kwargs)
            _args["file_extension"] = ".wav"

            result = super().convert(temp_path, **_args)
        finally:
            os.unlink(temp_path)

        return result


class ImageConverter(DocumentConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        import easyocr

        # Bail if not a XLSX
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".jpg", ".jpeg", ".png"]:
            return None

        ocr_min_confidence = kwargs.get("ocr_min_confidence", 0.25)

        md_content = ""

        # Add metadata
        metadata = self._get_metadata(local_path)
        if metadata:
            for f in [
                "Title",
                "Caption",
                "Description",
                "Keywords",
                "Artist",
                "DateTimeOriginal",
                "CreateDate",
                "GPSPosition",
            ]:
                if f in metadata:
                    md_content += f"{f}: {metadata[f]}\n"

        # Try describing the image with GPTV
        mlm_client = kwargs.get("mlm_client")
        if mlm_client is not None:
            md_content += (
                "\n# Description:\n"
                + self._get_mlm_description(local_path, extension, mlm_client, prompt=kwargs.get("mlm_prompt")).strip()
                + "\n"
            )

        image = PIL.Image.open(local_path)
        # Remove transparency
        if image.mode in ("RGBA", "P"):
            image = image.convert("RGB")

        reader = easyocr.Reader(["en"])  # specify the language(s)
        output = reader.readtext(np.array(image))  # local_path)
        # The output is a list of tuples, each containing the coordinates of the text and the text itself.
        # We join all the text pieces together to get the final text.
        ocr_text = " "
        for item in output:
            if item[2] >= ocr_min_confidence:
                ocr_text += item[1] + " "
        ocr_text = ocr_text.strip()

        if len(ocr_text) > 0:
            md_content += "\n# Text detected by OCR:\n" + ocr_text

        return DocumentConverterResult(
            title=None,
            text_content=md_content,
        )

    def _get_metadata(self, local_path):
        exiftool = shutil.which("exiftool")
        if not exiftool:
            return None
        else:
            try:
                result = subprocess.run([exiftool, "-json", local_path], capture_output=True, text=True).stdout
                return json.loads(result)[0]
            except Exception:
                return None

    def _get_mlm_description(self, local_path, extension, client, prompt=None):
        if prompt is None or prompt.strip() == "":
            prompt = "Write a detailed caption for this image."

        sys.stderr.write(f"MLM Prompt:\n{prompt}\n")

        data_uri = ""
        with open(local_path, "rb") as image_file:
            content_type, encoding = mimetypes.guess_type("_dummy" + extension)
            if content_type is None:
                content_type = "image/jpeg"
            image_base64 = base64.b64encode(image_file.read()).decode("utf-8")
            data_uri = f"data:{content_type};base64,{image_base64}"

            messages = [
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": prompt},
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": data_uri,
                            },
                        },
                    ],
                }
            ]

            return client(messages)


class ImageDoclingConverter(DoclingConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a supported image type
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".jpg", ".jpeg", ".png", ".tiff", ".bmp"]:
            return None
        return super().convert(local_path, **kwargs)


class MarkdownConverter(DocumentConverter):
    def convert(self, local_path, **kwargs) -> Union[None, DocumentConverterResult]:
        # Bail if not a Markdown file
        extension = kwargs.get("file_extension", "")
        if extension.lower() not in [".md", ".markdown"]:
            return None

        text_content = ""
        with open(local_path, "rt") as fh:
            text_content = fh.read()

        # Try to extract title from the first heading
        title = None
        # Look for the first markdown heading (# Title)
        heading_match = re.search(r"^#\s+(.+?)$", text_content, re.MULTILINE)
        if heading_match:
            title = heading_match.group(1).strip()

        return DocumentConverterResult(
            title=title,
            text_content=text_content,
        )


class FileConversionException(BaseException):
    pass


class UnsupportedFormatException(BaseException):
    pass


class FileConverterOptions(BaseModel):
    text_converter: DocumentConverter = Field(default_factory=PlainTextConverter)
    markdown_converter: DocumentConverter = Field(default_factory=MarkdownConverter)
    html_converter: DocumentConverter = Field(default_factory=HtmlConverter)
    wikipedia_converter: DocumentConverter = Field(default_factory=WikipediaConverter)
    youtube_converter: DocumentConverter = Field(default_factory=YouTubeConverter)
    docx_converter: DocumentConverter = Field(default_factory=DocxConverter)
    xlsx_converter: DocumentConverter = Field(default_factory=XlsxConverter)
    pptx_converter: DocumentConverter = Field(default_factory=PptxConverter)
    wav_converter: DocumentConverter = Field(default_factory=WavConverter)
    mp3_converter: DocumentConverter = Field(default_factory=Mp3Converter)
    image_converter: DocumentConverter = Field(default_factory=ImageConverter)
    pdf_converter: DocumentConverter = Field(default_factory=PdfMinerConverter)


class FileConverter:
    """(In preview) An extremely simple text-based document reader, suitable for LLM use.
    This reader will convert common file-types or webpages to Markdown."""

    def __init__(
        self,
        requests_session: Optional[requests.Session] = None,
        mlm_client: Optional[Any] = None,
        file_converter_options: Optional[FileConverterOptions] = None,
    ):
        if requests_session is None:
            self._requests_session = requests.Session()
        else:
            self._requests_session = requests_session

        self._mlm_client = mlm_client

        self._page_converters: List[DocumentConverter] = []

        if file_converter_options is None:
            file_converter_options = FileConverterOptions()

        # Register converters for successful browsing operations
        # Later registrations are tried first / take higher priority than earlier registrations
        # To this end, the most specific converters should appear below the most generic converters
        self.register_page_converter(file_converter_options.text_converter)
        self.register_page_converter(file_converter_options.markdown_converter)
        self.register_page_converter(file_converter_options.html_converter)
        self.register_page_converter(file_converter_options.wikipedia_converter)
        self.register_page_converter(file_converter_options.youtube_converter)
        self.register_page_converter(file_converter_options.docx_converter)
        self.register_page_converter(file_converter_options.xlsx_converter)
        self.register_page_converter(file_converter_options.pptx_converter)
        self.register_page_converter(file_converter_options.wav_converter)
        self.register_page_converter(file_converter_options.mp3_converter)
        self.register_page_converter(file_converter_options.image_converter)
        self.register_page_converter(file_converter_options.pdf_converter)

    def convert(self, source, **kwargs):
        """
        Args:
            source (str): can be a string representing a path or url, or a requests.response object
            **kwargs (dict, optional): additional options to pass to the converters
        """

        # Local path or url
        if isinstance(source, str):
            if source.startswith("http://") or source.startswith("https://") or source.startswith("file://"):
                return self.convert_url(source, **kwargs)
            else:
                return self.convert_local(source, **kwargs)
        # Request response
        elif isinstance(source, requests.Response):
            return self.convert_response(source, **kwargs)

    def convert_local(self, path, **kwargs):
        # Prepare a list of extensions to try (in order of priority)
        ext = kwargs.get("file_extension")
        extensions = [ext] if ext is not None else []

        # Get extension alternatives from the path and puremagic
        base, ext = os.path.splitext(path)
        self._append_ext(extensions, ext)
        self._append_ext(extensions, self._guess_ext_magic(path))

        # Convert
        return self._convert(path, extensions, **kwargs)

    def convert_url(self, url, **kwargs):
        # Send a HTTP request to the URL
        response = self._requests_session.get(url, stream=True)
        response.raise_for_status()
        return self.convert_response(response, **kwargs)

    def convert_response(self, response, **kwargs):
        # Prepare a list of extensions to try (in order of priority)
        ext = kwargs.get("file_extension")
        extensions = [ext] if ext is not None else []

        # Guess from the mimetype
        content_type = response.headers.get("content-type", "").split(";")[0]
        self._append_ext(extensions, mimetypes.guess_extension(content_type))

        # Read the content disposition if there is one
        content_disposition = response.headers.get("content-disposition", "")
        m = re.search(r"filename=([^;]+)", content_disposition)
        if m:
            base, ext = os.path.splitext(m.group(1).strip("\"'"))
            self._append_ext(extensions, ext)

        # Read from the extension from the path
        base, ext = os.path.splitext(urlparse(response.url).path)
        self._append_ext(extensions, ext)

        # Save the file locally to a temporary file. It will be deleted before this method exits
        handle, temp_path = tempfile.mkstemp(suffix=ext)
        fh = os.fdopen(handle, "wb")
        result = None
        try:
            # Download the file
            for chunk in response.iter_content(chunk_size=512):
                fh.write(chunk)
            fh.close()

            # Use puremagic to check for more extension options
            self._append_ext(extensions, self._guess_ext_magic(temp_path))

            # Convert
            result = self._convert(temp_path, extensions, url=response.url, **kwargs)

        # Clean up
        finally:
            try:
                fh.close()
            except Exception:
                pass
            os.unlink(temp_path)

        return result

    def _convert(self, local_path, extensions, **kwargs):
        error_trace = ""
        for ext in extensions:
            for converter in self._page_converters:
                _kwargs = copy.deepcopy(kwargs)
                _kwargs.update({"file_extension": ext})

                # Copy any additional global options
                if "mlm_client" not in _kwargs and self._mlm_client is not None:
                    _kwargs["mlm_client"] = self._mlm_client

                # If we hit an error log it and keep trying
                res = None
                try:
                    res = converter.convert(local_path, **_kwargs)
                except Exception:
                    logger.exception(f"Error converting {local_path} with {converter.__class__.__name__}")
                    error_trace = ("\n\n" + traceback.format_exc()).strip()

                if res is not None:
                    # Normalize the content
                    res.text_content = "\n".join([line.rstrip() for line in re.split(r"\r?\n", res.text_content)])
                    res.text_content = re.sub(r"\n{3,}", "\n\n", res.text_content)

                    # Todo
                    return res

        # If we got this far without success, report any exceptions
        if len(error_trace) > 0:
            raise FileConversionException(
                f"Could not convert '{local_path}' to Markdown. File type was recognized as {extensions}. While converting the file, the following error was encountered:\n\n{error_trace}"
            )

        # Nothing can handle it!
        raise UnsupportedFormatException(
            f"Could not convert '{local_path}' to Markdown. The formats {extensions} are not supported."
        )

    def _append_ext(self, extensions, ext):
        """Append a unique non-None, non-empty extension to a list of extensions."""
        if ext is None:
            return
        ext = ext.strip()
        if ext == "":
            return
        if ext not in extensions:
            extensions.append(ext)

    def _guess_ext_magic(self, path):
        """Use puremagic (a Python implementation of libmagic) to guess a file's extension based on the first few bytes."""
        # Use puremagic to guess
        try:
            guesses = puremagic.magic_file(path)
            if len(guesses) > 0:
                ext = guesses[0].extension.strip()
                if len(ext) > 0:
                    return ext
        except FileNotFoundError:
            pass
        except IsADirectoryError:
            pass
        except PermissionError:
            pass
        return None

    def register_page_converter(self, converter: DocumentConverter) -> None:
        """Register a page text converter."""
        self._page_converters.insert(0, converter)



import contextlib
import os
import shutil
import sys
import tempfile
from pathlib import Path

import testbook
import yaml
from omegaconf import DictConfig

import tapeagents.observe
from examples import delegate_stack
from examples.data_science import data_science
from examples.optimize import optimize
from examples.tape_improver import tape_improver


@contextlib.contextmanager
def run_test_in_tmp_dir(test_name: str):
    """Copy test resources to a temporary directory and run the test there"""
    cur_dir = os.getcwd()
    tmpdir = tempfile.mkdtemp()
    test_data_dir = Path(f"{test_name}").resolve()
    os.chdir(tmpdir)
    shutil.copytree(test_data_dir, tmpdir, dirs_exist_ok=True)
    # force creation of SQLite tables
    tapeagents.observe._checked_sqlite = False
    yield tmpdir
    os.chdir(cur_dir)


@contextlib.contextmanager
def run_in_tmp_dir_to_make_test_data(test_group, test_name: str, keep_llm_cache=False):
    cur_dir = os.getcwd()
    tmpdir = tempfile.mkdtemp()
    os.chdir(tmpdir)
    try:
        yield
    # find all non-directory files that got created
    finally:
        created_files = []
        for root, _, files in os.walk(tmpdir):
            for file in files:
                # For most of the code we test in TapeAgents we create ReplayLLM
                # that looks up prompts and outputs in the SQLite database. For this
                # reason, by default we don't save the LLM cache files. If you want
                # make test data for a Jupyter notebook, you can use the keep_llm_cache
                # to save the LLM cache files.
                if file.startswith("llm_cache") and not keep_llm_cache:
                    continue
                created_files.append(os.path.relpath(os.path.join(root, file), tmpdir))
        cp_source = " ".join(f"$TMP/{f}" for f in created_files)
        test_data_dir = f"tests/{test_group}/res/{test_name}"
        print("Saved test data to ", tmpdir)
        print("To update test data, run these commands:")
        print(f"mkdir {test_data_dir}")
        print(f"TMP={tmpdir}; cp {cp_source} {test_data_dir}")
    os.chdir(cur_dir)


if __name__ == "__main__":
    match sys.argv[1:]:
        case ["delegate_stack"]:
            with run_in_tmp_dir_to_make_test_data("examples", "delegate_stack"):
                delegate_stack.main()
        case ["intro_notebook"]:
            intro_notebook_path = Path("intro.ipynb").resolve()
            assets_path = Path("assets").resolve()
            with run_in_tmp_dir_to_make_test_data("examples", "intro_notebook", keep_llm_cache=True):
                shutil.copytree(assets_path, Path("assets"))
                with testbook.testbook(intro_notebook_path) as tb:
                    tb.inject(
                        """
                        from tapeagents import llms
                        llms._force_cache = True
                        """,
                        before=0,
                    )
                    tb.execute()
        case ["tape_improver"]:
            with run_in_tmp_dir_to_make_test_data("examples", "tape_improver"):
                tape_improver.main("run improver")
        case ["data_science"]:
            with run_in_tmp_dir_to_make_test_data("examples", "data_science"):
                data_science.main(studio=False)
        case ["optimize"]:
            models = [
                # "gpt-3.5-turbo",
                "gpt-4o-mini",
                # "llama-3.3-70b-instruct"
            ]
            for model in models:
                with open(f"tests/examples/res/optimize/{model}/config.yaml") as f:
                    cfg = DictConfig(yaml.safe_load(f))
                with run_in_tmp_dir_to_make_test_data("examples", f"optimize/{model}", keep_llm_cache=True):
                    optimize.run(cfg)
        case _:
            raise Exception(
                "Usage: python -m tests.make_test_data [delegate_stack | intro_notebook | tape_improver | data_science | optimize]"
            )



from __future__ import annotations

import json
from typing import Annotated, Literal, TypeAlias

from pydantic import Field

from tapeagents.agent import Agent
from tapeagents.core import (
    Action,
    FinalStep,
    Prompt,
    Tape,
    Thought,
)
from tapeagents.llms import LLM, LLMStream, TrainableLLM
from tapeagents.nodes import CallSubagent
from tapeagents.team import Chain
from tapeagents.view import Call, Respond, TapeViewStack

EXAMPLE_TEXT = """I am a text with some verbs like running, jumping, and swimming."""

FIND_VERBS_MESSAGE = """Your job is to find all verbs in the following text:

{text}

You must output JSON like this:

{{
    "verbs": ["verb1", "verb2", ...]
}} 
Do not output any text before or after JSON."""

FIND_NOUNS_MESSAGE = """Your job is to find all nouns in the following text:

{text}

You must output JSON like this:
    
{{
    "nouns": ["noun1", "noun2", ...]
}}

Do not output any text before or after JSON. Use the infinitive form for each verb in your output.
"""

FILTER_IRREGULAR_MESSAGE = """Your job is to filter out irregular verbs from this list of verbs you received below. 

{verbs}

You must output JSON like this:

{{
    "verbs": ["verb1", "verb2", ...]

}}

Do not output any text before or after JSON. Use the infinitive form for each verb in your output.
"""

PRESENT_RESULTS_MESSAGE = """You are a linguist who has done some analysis of the text. 

You found some nouns: {nouns}
And you found some irregular verbs: {irregural_verbs}

Write a nice message discussing the style of the text based on the information you found."""


class InputText(Thought):
    """Thought that the agent should use to pass text as an input to a worker"""

    kind: Literal["input_text"] = "input_text"
    text: str


class AllVerbs(Thought):
    kind: Literal["all_verbs"] = "all_verbs"
    verbs: list[str]


class AllNouns(Thought):
    kind: Literal["all_nouns"] = "all_nouns"
    nouns: list[str]


class IrregularVerbs(Thought):
    kind: Literal["irregular_verbs"] = "irregular_verbs"
    verbs: list[str]


class PresentResults(Action):
    kind: Literal["present_results"] = "present_results"
    content: str


class LastStep(FinalStep):
    kind: Literal["last_step"] = "last_step"


KnownStep: TypeAlias = Annotated[
    InputText | AllVerbs | AllNouns | IrregularVerbs | Call | Respond | LastStep | PresentResults,
    Field(discriminator="kind"),
]
ExampleTape = Tape[str, KnownStep]


class FindVerbs(Agent[ExampleTape]):
    def make_prompt(self, tape: ExampleTape):
        return Prompt.from_user_message(self.templates["default"].format(text=tape.context))

    def generate_steps(self, tape: ExampleTape, llm_stream: LLMStream):
        yield AllVerbs.model_validate_json(llm_stream.get_text())
        yield Respond(copy_output=True)


class FilterIrregular(Agent[ExampleTape]):
    def make_prompt(self, tape: ExampleTape):
        assert isinstance(tape.steps[-1], AllVerbs)
        return Prompt.from_user_message(self.template.format(verbs=json.dumps(tape.steps[-1].verbs)))

    def generate_steps(self, tape: ExampleTape, llm_stream: LLMStream):
        yield IrregularVerbs.model_validate_json(llm_stream.get_text())
        yield Respond(copy_output=True)


class FindNouns(Agent[ExampleTape]):
    def make_prompt(self, tape: ExampleTape):
        return Prompt.from_user_message(self.template.format(text=tape.context))

    def generate_steps(self, tape: ExampleTape, llm_stream: LLMStream):
        yield AllNouns.model_validate_json(llm_stream.get_text())
        yield Respond(copy_output=True)


class Linguist(Chain[ExampleTape]):
    """Analyze the style of the text based on the nouns and irregular verbs that it contains.

    This version shows how you a Chain agent can make its own prompts and generate its own steps.

    """

    @classmethod
    def create(cls, llm: LLM):  # type: ignore
        return super().create(
            llms=llm,
            nodes=[
                CallSubagent(agent=FindNouns.create(llms=llm, templates=FIND_NOUNS_MESSAGE)),
                CallSubagent(
                    agent=Chain.create(
                        name="FindIrregularVerbs",
                        nodes=[
                            CallSubagent(agent=FindVerbs.create(llm, templates=FIND_VERBS_MESSAGE)),
                            CallSubagent(
                                agent=FilterIrregular.create(llm, templates=FILTER_IRREGULAR_MESSAGE),
                                inputs=(-1,),
                            ),
                        ],
                    ),
                ),
            ],
            templates=PRESENT_RESULTS_MESSAGE,
        )

    def make_prompt(self, tape: ExampleTape) -> Prompt:
        state = TapeViewStack.compute(tape)
        if "all_nouns" in state.top.steps_by_kind and "irregular_verbs" in state.top.steps_by_kind:
            (nouns,) = state.top.steps_by_kind["all_nouns"]
            irregular_verbs = state.top.steps_by_kind["irregular_verbs"][0]
            assert isinstance(nouns, AllNouns)
            assert isinstance(irregular_verbs, IrregularVerbs)
            return Prompt.from_user_message(
                self.template.format(
                    nouns=json.dumps(nouns.nouns),
                    irregural_verbs=json.dumps(irregular_verbs.verbs),
                )
            )
        else:
            return Prompt()

    def generate_steps(self, tape: ExampleTape, llm_stream: LLMStream):
        if llm_stream:
            yield PresentResults(content=llm_stream.get_text())
        else:
            yield from super().generate_steps(tape, llm_stream)


class PresentAnalysis(Agent[ExampleTape]):
    def make_prompt(self, tape: ExampleTape):
        assert isinstance(tape.steps[-2], AllNouns)
        assert isinstance(tape.steps[-1], IrregularVerbs)
        return Prompt.from_user_message(
            self.template.format(
                nouns=json.dumps(tape.steps[-2].nouns),
                irregural_verbs=json.dumps(tape.steps[-1].verbs),
            )
        )

    def generate_steps(self, tape: ExampleTape, llm_stream: LLMStream):
        yield PresentResults(content=llm_stream.get_text())
        yield Respond(copy_output=True)


def make_analyze_text_chain(llm: LLM):
    """
    The agent that analyzes the text for nouns and irregular verbs and then presents the results.
    """
    return Chain.create(
        name="Linguist",
        nodes=[
            CallSubagent(agent=FindNouns.create(llms=llm, templates=FIND_NOUNS_MESSAGE)),
            CallSubagent(
                agent=Chain.create(
                    name="FindIrregularVerbs",
                    nodes=[
                        CallSubagent(agent=FindVerbs.create(llm, templates=FIND_VERBS_MESSAGE)),
                        CallSubagent(
                            agent=FilterIrregular.create(llm, templates=FILTER_IRREGULAR_MESSAGE),
                            inputs=(-1,),
                        ),
                    ],
                ),
            ),
            CallSubagent(agent=PresentAnalysis.create(llm, templates=PRESENT_RESULTS_MESSAGE), inputs=(-2, -1)),
        ],
    )


def main():
    llama = TrainableLLM(
        base_url="https://api.together.xyz",
        model_name="meta-llama/Meta-Llama-3-70B-Instruct-Turbo",
        tokenizer_name="meta-llama/Meta-Llama-3-70B-Instruct",
        parameters=dict(temperature=0.7, max_tokens=512),
        use_cache=True,
    )
    with open("llm.json", "w") as f:
        json.dump(llama.model_dump(), f, indent=2)
    tape = ExampleTape(context=EXAMPLE_TEXT)
    agent1 = Linguist.create(llama)
    with open("start_tape.json", "w") as f:
        json.dump(tape.model_dump(), f, indent=2)
    agent1 = Linguist.create(llama)
    tape1 = agent1.run(tape).get_final_tape()
    print(tape1.model_dump_json(indent=2))
    with open("tape1.json", "w") as f:
        json.dump(tape1.model_dump(), f, indent=2)
    agent2 = make_analyze_text_chain(llama)
    tape2 = agent2.run(tape).get_final_tape()
    print(tape2.model_dump_json(indent=2))
    with open("tape2.json", "w") as f:
        json.dump(tape2.model_dump(), f, indent=2)


if __name__ == "__main__":
    main()



"""
Core data structures for the tape agents framework.
"""

from __future__ import annotations

import datetime
import json
from typing import Any, Generic, Iterable, Iterator, List, Literal, Type, TypeAlias, TypeVar
from uuid import uuid4

import litellm
from pydantic import BaseModel, Field, SerializeAsAny
from typing_extensions import Self


class TrainingText(BaseModel):
    """
    Training text instance used to finetune a language model.

    Attributes:
        text (str): The full text of the training instance.
        n_predicted (int): The number of predicted characters in the text.
        reward (float): The reward associated with the training instance. Defaults to 0.0.
        logprobs (List[float]): A list of log probabilities of the completion tokens from the assistant model.
        ref_logprobs (List[float]): A list of reference log probabilities of the completion tokens from the reference model.
        input_ids (List[int]): The tokenized input ids of the text.
        labels (List[int]): The tokenized labels of the text (i.e., masked token ids for the prompt and regular token ids for the prediction).
        group_id (str, optional): ID of the group. It is used by the RL finetuning script to normalize rewards.
        prompt_text (str): Portion of the text that serves as the prompt (i.e., the text excluding the predicted characters).
        output_text (str): Portion of the text that represents the predicted output (i.e., the last n_predicted characters).
    """

    text: str
    n_predicted: int
    reward: float = 0.0
    logprobs: List[float] = Field(default_factory=list)
    ref_logprobs: List[float] = Field(default_factory=list)
    input_ids: List[int] = Field(default_factory=list)
    labels: List[int] = Field(default_factory=list)
    group_id: str | None = None
    metadata: dict = Field(default_factory=dict)

    @property
    def prompt_text(self) -> str:
        return self.text[: -self.n_predicted]

    @property
    def output_text(self) -> str:
        return self.text[-self.n_predicted :]


class StepMetadata(BaseModel):
    """
    StepMetadata is a model that represents metadata for a tape step.

    Attributes:
        id (str): A unique identifier for the step, generated using uuid4.
        prompt_id (str): An identifier for the llm prompt used to produce the step.
        node (str): The node that produced the step.
        agent (str): The agent that produced the step.
        other (dict[str, Any]): A dictionary to store additional metadata related to the step.
    """

    id: str = Field(default_factory=lambda: str(uuid4()))
    prompt_id: str = ""
    node: str = ""
    agent: str = ""
    llm: str = ""
    other: dict[str, Any] = Field(default_factory=dict)


class Step(BaseModel):
    """
    Base class representing a step in a tape.

    Attributes:
        metadata (StepMetadata): Metadata associated with the step.
        kind (Literal["define_me"]): A placeholder value indicating the kind of step.
                                     This should be overwritten in subclasses.
    """

    metadata: StepMetadata = Field(default_factory=StepMetadata)
    kind: Literal["define_me"] = "define_me"  # This is a placeholder value, it should be overwritten in subclasses

    def llm_dict(self) -> dict[str, Any]:
        """Dumps the step data as dictionary, excluding the metadata."""
        return self.model_dump(exclude_none=True, exclude={"metadata"})

    def llm_view(self, indent: int | None = 2) -> str | list[dict]:
        """Returns a JSON string representation of the step data, excluding the metadata."""
        return json.dumps(self.llm_dict(), indent=indent, ensure_ascii=False)

    @classmethod
    def get_kind(cls) -> str:
        """Returns the default value of the 'kind' field."""
        return cls.model_fields["kind"].default


class PartialStep(BaseModel):
    """
    Wrap your step as partial step to indicate that it is not finished yet.

    Attributes:
        step (Step): An instance of the Step class representing the step details.
    """

    step: Step


class Observation(Step):
    """
    Base class representing an observation in a tape.
    """

    def short_view(self, max_chars=100) -> str:
        """Returns a short string representation of the observation when the tape needs to be trimmed."""
        return self.llm_view()


class Error(Step):
    """
    Base class representing an error in a tape.
    """

    pass


class AgentStep(Step):
    """
    Base class representing a step produced by an agent.
    """

    pass


class Thought(AgentStep):
    """
    Base class representing an agent's thought in a tape.
    """

    pass


class ControlFlow(Thought):
    """
    Base class representing a flow control step in a tape.
    """

    pass


class Action(AgentStep):
    """
    Base class representing an agent's action in a tape.
    """

    @classmethod
    def description(cls) -> str:
        return f"{cls.__name__} - {cls.__doc__ or '[no description]'}"


class LLMOutputParsingFailureAction(Action, Error):
    """
    Represents an action that is produced automatically when the LLM output parsing fails.

    Attributes:
        kind (Literal["llm_output_parsing_failure_action"]): A constant string indicating the type of action.
        error (str): A description of the error that occurred during parsing.
        llm_output (str): The raw output from the LLM that could not be parsed.
    """

    kind: Literal["llm_output_parsing_failure_action"] = "llm_output_parsing_failure_action"
    error: str
    llm_output: str


class StopStep(Action):
    """
    Action that stops orchestrator loop.
    """

    pass


class FinalStep(StopStep):
    """
    Action that stops orchestrator loop with a reason.

    Attributes:
        kind (Literal["final_step"]): A constant string indicating the type of action.
        reason (str): The reason for stopping the orchestr
    """

    kind: Literal["final_step"] = "final_step"
    reason: str = ""


class FinalObservation(Observation, StopStep):
    kind: Literal["final_observation"] = "final_observation"


class SetNextNode(ControlFlow):
    """
    Action that sets the next node to run in the current agent.

    Attributes:
        kind (Literal["set_next_node"]): A constant string indicating the type of action.
        next_node (str): The name of the next node to run.
    """

    kind: Literal["set_next_node"] = "set_next_node"
    next_node: str


class Pass(Thought):
    """
    Action that does nothing.
    """

    kind: Literal["pass"] = "pass"


class Call(ControlFlow):
    """
    Action that calls another agent.

    Attributes:
        kind (Literal["call"]): A constant string indicating the type of action.
        content (str): The content passed with the call to the agent.
        agent_name (str): The name of the agent to call.
    """

    kind: Literal["call"] = "call"
    content: str = ""
    agent_name: str


class Respond(ControlFlow):
    """
    Action that returns a response to the top-level agent after processing the call.

    Attributes:
        kind (Literal["respond"]): A constant string indicating the type of action.
        content (str): The content of the response.
        copy_output (bool): Indicates whether the last step before this one should be copied to the top-level agent.
    """

    content: str = ""
    kind: Literal["respond"] = "respond"
    copy_output: bool = False


StepType = TypeVar("StepType", bound=Action | Observation | Thought)
"""Type variable for step types."""


class TapeMetadata(BaseModel):
    """
    TapeMetadata is a model that represents metadata information for a tape.

    Attributes:
        id (str): A unique identifier for the tape, generated by default using uuid4.
        parent_id (str, optional): An optional identifier for the parent tape.
        author (str, optional): An optional name of the author of the tape.
        author_tape_id (str, optional): An optional identifier for the author's tape.
        n_added_steps (int, optional): The number of steps added to the tape in the last agent run, default is 0.
        error (Any, optional): An optional field to store any errors occured during the last agent run.
        result (Any, optional): Optional field to store the result associated with the tape, default is an empty dictionary.
    """

    id: str = Field(default_factory=lambda: str(uuid4()))
    parent_id: str | None = None
    author: str | None = None
    author_tape_id: str | None = None
    n_added_steps: int = 0
    error: Any | None = None
    result: Any = {}


ContextType = TypeVar("ContextType")
"""Type variable for context types."""


class Tape(BaseModel, Generic[ContextType, StepType]):
    """
    Tape class represents a sequence of steps produced by agents and environments with associated metadata.

    Supports iteration, slicing, concatenation, and appending of steps.

    Attributes:
        metadata (TapeMetadata): Metadata associated with the tape.
        context (ContextType, optional): Context information for the tape.
        steps (List[StepType]): List of steps in the tape.
    """

    metadata: TapeMetadata = Field(default_factory=TapeMetadata)
    context: ContextType | None = None
    steps: List[StepType] = Field(default_factory=list)

    def __iter__(self) -> Iterator[StepType]:  # type: ignore
        return iter(self.steps)

    def __len__(self) -> int:
        return len(self.steps)

    def __getitem__(self, key: int | slice) -> StepType | Self:
        if isinstance(key, slice):  # cut and erase metadata
            return self.model_copy(update=dict(steps=self.steps[key.start : key.stop], metadata=TapeMetadata()))
        return self.steps[key]

    def __add__(self, tape: Self | Iterable[Step]) -> Self:
        """
        Concatenate two tapes or append list of steps to the tape
        """
        new_steps = tape.steps if isinstance(tape, Tape) else list(tape)
        return self.model_copy(
            update=dict(
                steps=self.steps + new_steps,
                metadata=TapeMetadata(n_added_steps=len(new_steps)),
            )
        )

    def append(self, step: StepType) -> Self:
        """
        Adds a step to the tape and creates new metadata (new tape id, etc...).
        """
        return self.model_copy(update=dict(steps=self.steps + [step], metadata=TapeMetadata()))

    def with_new_id(self) -> Self:
        """
        Creates a copy of the tape with new metadata.
        """
        return self.model_copy(update=dict(metadata=TapeMetadata()))


TapeType = TypeVar("TapeType", bound=Tape)
"""Type variable for tape types."""


class Prompt(BaseModel):
    """
    A class representing a LLM prompt with messages and tools.

    Attributes:
        id (str): A unique identifier for the prompt, generated by default.
        tools (list[dict], optional): A list of tools associated with the prompt, default is None.
        messages (list[dict]): A list of messages in the prompt, default is an empty list.
    """

    id: str = Field(default_factory=lambda: str(uuid4()))
    tools: list[dict] | None = None
    messages: list[dict] = Field(default_factory=list)
    token_ids: list[int] = Field(default_factory=list)
    response_format: dict | Type[BaseModel] | None = None

    @staticmethod
    def from_user_message(content: str) -> Prompt:
        """
        Creates a Prompt instance from a user message.

        Args:
            content (str): The content of the user message.

        Returns:
            Prompt: A Prompt instance with the user message.
        """
        return Prompt(messages=[{"role": "user", "content": content}])

    def __bool__(self) -> bool:
        return bool(self.messages)


LLMOutput: TypeAlias = litellm.utils.Message
"""Type alias for the output of the language model."""


class LLMCall(BaseModel):
    """
    LLMCall stores info about a call to a language model.

    Attributes:
        timestamp (str): The timestamp when the call was made, in ISO 8601 format.
        prompt (Prompt): The input prompt provided to the language model.
        output (LLMOutput): The output generated by the language model.
        prompt_length_tokens (int): The length of the prompt in tokens. Defaults to -1 if not set.
        output_length_tokens (int): The length of the output in tokens. Defaults to -1 if not set.
        cached (bool): Indicates whether the result was retrieved from cache.
    """

    timestamp: str = Field(default_factory=lambda: datetime.datetime.now().isoformat())
    prompt: Prompt
    output: LLMOutput
    prompt_length_tokens: int = -1
    output_length_tokens: int = -1
    cached: bool
    llm_info: dict = {}
    cost: float = 0
    logprobs: list[TokenLogprob] = Field(default_factory=list, exclude=True)


class TokenLogprob(BaseModel):
    logprob: float
    token_id: int


AnnotatorTape = Tape[TapeType, StepType]
"""Type alias for annotator tapes."""
AnnotatorTapeType = TypeVar("AnnotatorTapeType", bound=AnnotatorTape)
"""Type variable for annotator tape types."""


class AnnotationWithMetadata(BaseModel):
    annotation: Any
    annotator_tape_id: str


class Episode(BaseModel):
    """
    Auxiliary data structure for tape with annotations attached.
    Humans may want to look at a tape with one-or-many annotations attached to agent's actions.
    This is a data structure to store such tapes. We store the tape id for traceability.
    """

    tape: Tape
    annotator_tapes: dict[int, list[Tape]]
    obs_making_tapes: dict[int, Tape]

    def group_by_step(self) -> Iterator[tuple[Tape | None, Step, list[Tape]]]:
        for i, step in enumerate(self.tape.steps):
            yield (
                self.obs_making_tapes.get(i, None),
                step,
                self.annotator_tapes.get(i, []),
            )


class AgentEvent(BaseModel, Generic[TapeType]):
    """
    Event produced by the agent during the run.

    Can contain a step, a final tape with all new steps, or a partial step when used in streaming mode.
    Attributes are mutually exclusive.

    Attributes:
        step (SerializeAsAny[Step], optional): A step produced by the agent.
        partial_step (PartialStep, optional): A partial step produced by the agent.
        partial_tape (TapeType, optional): A partial tape produced by the agent.
        final_tape (TapeType, optional): A final tape produced by the agent.
    """

    step: SerializeAsAny[Step] | None = None
    partial_step: PartialStep | None = None
    partial_tape: TapeType | None = None
    final_tape: TapeType | None = None


ObservationMakerTapeType = TypeVar("ObservationMakerTapeType", bound=Tape)


class MakeObservation(Action, Generic[StepType]):
    kind: Literal["make_observation"] = "make_observation"
    new_observation: StepType

    def llm_dict(self) -> dict[str, Any]:
        """Dumps the step data as dictionary, excluding the metadata of the step itself and the metadata of the wrapped step"""
        obj = self.model_dump(exclude_none=True, exclude={"metadata"})
        del obj["new_observation"]["metadata"]
        return obj


def last_actions(tape: Tape) -> list[Action]:
    return [step for step in tape.steps[-tape.metadata.n_added_steps :] if isinstance(step, Action)]



from pathlib import Path
from typing import Literal, Optional

from pydantic import Field

from tapeagents.core import Action, Observation
from tapeagents.tools.base import Tool
from tapeagents.tools.converters import (
    FileConversionException,
    FileConverter,
    FileConverterOptions,
    UnsupportedFormatException,
)


def read_document(path: str, file_converter_options: Optional[FileConverterOptions] = None) -> tuple[str, str | None]:
    """Read a document, file or image and and convert it to Markdown."""
    try:
        text = ""
        error = None
        text = FileConverter(file_converter_options=file_converter_options).convert(path).text_content
    except UnsupportedFormatException as e:
        error = f"Failed to read document {path}: {e}"
    except FileConversionException as e:
        error = f"Failed to read document {path}: {e}"
    except Exception as e:
        error = f"Failed to read document {path}: {e}"
    return text, error


class DocumentObservation(Observation):
    kind: Literal["document_observation"] = "document_observation"
    text: str
    source: str | None = None  # optional for tape backwards compatibility
    error: str | None = None


class ReadLocalDocumentAction(Action):
    """
    Action that loads the document, file or image and converts it to Markdown.
    """

    kind: Literal["read_local_document_action"] = "read_local_document_action"
    path: str = Field(description="path of the document")


class DocumentReader(Tool):
    """
    Tool to read a document and convert it to Markdown.
    """

    action: type[Action] = ReadLocalDocumentAction
    observation: type[Observation] = DocumentObservation
    file_converter_options: Optional[FileConverterOptions] = None
    workspace_directory: str | None = None

    def execute_action(self, action: ReadLocalDocumentAction) -> DocumentObservation:
        if self.workspace_directory is not None:
            # Resolve the path safely within the workspace directory
            workspace_dir = Path(self.workspace_directory).resolve()
            path = Path(action.path)
            abs_path = (workspace_dir / path).resolve() if not path.is_absolute() else path.resolve()
            try:
                abs_path.relative_to(workspace_dir)
            except ValueError:
                return DocumentObservation(
                    text="", error="Access to files outside the workspace directory is not allowed"
                )
            workspace_path = abs_path
        else:
            workspace_path = Path(action.path)
        if not workspace_path.exists():
            return DocumentObservation(text="", error=f"File {action.path} not found in the workspace")
        text, error = read_document(str(workspace_path), self.file_converter_options)
        return DocumentObservation(text=text, source=str(action.path), error=error)


class ListDocumentsAction(Action):
    """
    Action that lists all documents in the workspace directory.
    """

    kind: Literal["list_documents_action"] = "list_documents_action"  # type: ignore


class ListDocumentsObservation(Observation):
    """
    Observation that lists all documents in the workspace directory.
    """

    kind: Literal["list_documents_observation"] = "list_documents_observation"  # type: ignore
    documents: list[str] = Field(description="List of documents in the workspace directory")
    error: str | None = None


class ListDocuments(Tool):
    """
    Tool to list all documents in the workspace directory.
    """

    action: type[Action] = ListDocumentsAction
    observation: type[Observation] = ListDocumentsObservation
    workspace_directory: str

    def execute_action(self, action: ListDocumentsAction) -> ListDocumentsObservation:
        try:
            documents = self.list_documents()
            return ListDocumentsObservation(documents=documents)
        except Exception as e:
            return ListDocumentsObservation(documents=[], error=str(e))

    def list_documents(self) -> list[str]:
        """
        List all documents in the workspace directory.
        """
        import os

        documents = []
        os.makedirs(self.workspace_directory, exist_ok=True)
        for root, _, files in os.walk(self.workspace_directory):
            for file in files:
                abs_path = os.path.join(root, file)
                rel_path = os.path.relpath(abs_path, self.workspace_directory)
                documents.append(rel_path)
        return documents



import os
import sqlite3
import sys

_llm_calls = "LLMCalls"


def add_columns_to_db(db_path):
    conn = sqlite3.connect(db_path)
    c = conn.cursor()
    c.execute(f"PRAGMA table_info{_llm_calls}")
    columns = [row[1] for row in c.fetchall()]
    if "llm_info" not in columns:
        c.execute(f"ALTER TABLE {_llm_calls} ADD COLUMN llm_info TEXT DEFAULT '{{}}'")
    if "cost" not in columns:
        c.execute(f"ALTER TABLE {_llm_calls} ADD COLUMN cost REAL DEFAULT 0.0")
    conn.commit()
    conn.close()


def main():
    dirname = sys.argv[1]
    for root, dirs, files in os.walk(dirname):
        for file in files:
            if file.endswith(".sqlite"):
                db_path = os.path.join(root, file)
                print(db_path)
                add_columns_to_db(db_path)


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Usage: python -m tests.add_sqlite_columns <dirname>"
    main()



# derived from https://github.com/pyparsing/pyparsing/blob/master/examples/fourFn.py

from __future__ import division

import json
import math
import operator
import re
from typing import Any, Literal

from pydantic import Field
from pyparsing import (
    CaselessLiteral,
    Combine,
    Forward,
    Group,
    Literal as ParsingLiteral,
    Optional,
    Word,
    ZeroOrMore,
    alphas,
    nums,
    oneOf,
)

from tapeagents.core import Action, Observation
from tapeagents.tools.base import Tool


def cmp(a, b):
    return (a > b) - (a < b)


class NumericStringParser(object):
    def pushFirst(self, strg, loc, toks):
        self.exprStack.append(toks[0])

    def pushUMinus(self, strg, loc, toks):
        if toks and toks[0] == "-":
            self.exprStack.append("unary -")

    def __init__(self):
        """
        expop   :: '^'
        multop  :: '*' | '/'
        addop   :: '+' | '-'
        integer :: ['+' | '-'] '0'..'9'+
        atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
        factor  :: atom [ expop factor ]*
        term    :: factor [ multop factor ]*
        expr    :: term [ addop term ]*
        """
        point = ParsingLiteral(".")
        e = CaselessLiteral("E")
        fnumber = Combine(
            Word("+-" + nums, nums) + Optional(point + Optional(Word(nums))) + Optional(e + Word("+-" + nums, nums))
        )
        ident = Word(alphas, alphas + nums + "_$")
        plus = ParsingLiteral("+")
        minus = ParsingLiteral("-")
        mult = ParsingLiteral("*")
        div = ParsingLiteral("/")
        lpar = ParsingLiteral("(").suppress()
        rpar = ParsingLiteral(")").suppress()
        addop = plus | minus
        multop = mult | div
        expop = ParsingLiteral("^")
        pi = CaselessLiteral("PI")
        expr = Forward()
        atom = (
            (Optional(oneOf("- +")) + (ident + lpar + expr + rpar | pi | e | fnumber).setParseAction(self.pushFirst))
            | Optional(oneOf("- +")) + Group(lpar + expr + rpar)
        ).setParseAction(self.pushUMinus)
        # by defining exponentiation as "atom [ ^ factor ]..." instead of
        # "atom [ ^ atom ]...", we get right-to-left exponents, instead of left-to-right
        # that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor << atom + ZeroOrMore((expop + factor).setParseAction(self.pushFirst))
        term = factor + ZeroOrMore((multop + factor).setParseAction(self.pushFirst))
        expr << term + ZeroOrMore((addop + term).setParseAction(self.pushFirst))
        # addop_term = ( addop + term ).setParseAction( self.pushFirst )
        # general_term = term + ZeroOrMore( addop_term ) | OneOrMore( addop_term)
        # expr <<  general_term
        self.bnf = expr
        # map operator symbols to corresponding arithmetic operations
        epsilon = 1e-12
        self.opn = {"+": operator.add, "-": operator.sub, "*": operator.mul, "/": operator.truediv, "^": operator.pow}
        self.fn = {
            "sin": math.sin,
            "cos": math.cos,
            "tan": math.tan,
            "exp": math.exp,
            "abs": abs,
            "trunc": lambda a: int(a),
            "round": round,
            "sgn": lambda a: abs(a) > epsilon and cmp(a, 0) or 0,
        }

    def evaluateStack(self, s):
        op = s.pop()
        if op == "unary -":
            return -self.evaluateStack(s)
        if op in "+-*/^":
            op2 = self.evaluateStack(s)
            op1 = self.evaluateStack(s)
            return self.opn[op](op1, op2)
        elif op == "PI":
            return math.pi  # 3.1415926535
        elif op == "E":
            return math.e  # 2.718281828
        elif op in self.fn:
            return self.fn[op](self.evaluateStack(s))
        elif op[0].isalpha():
            return 0
        else:
            return float(op)

    def eval(self, num_string, parseAll=True):
        self.exprStack = []
        self.bnf.parseString(num_string, parseAll)
        val = self.evaluateStack(self.exprStack[:])
        return val


def calculate(expr: str, values_dict: dict[str, Any]) -> str:
    pairs = values_dict.items()
    pairs = sorted(pairs, key=lambda x: len(x[0]), reverse=True)  # substitute longest vars first
    for k, v in pairs:
        if isinstance(v, str):
            expr = re.sub(k, "'{}'".format(v), expr)
        else:
            expr = re.sub(k, str(v), expr)
    try:
        result = NumericStringParser().eval(expr)
    except Exception:
        raise ValueError(f"Error evaluating expression: {expr}")
    try:
        str_result = json.dumps(result)
    except Exception:
        str_result = str(result)
    return str_result


class UseCalculatorAction(Action):
    """
    Action to use calculator to find the new fact. This python math expression uses only the fact names from the previous steps and constants. The expression should be a single line. You can use exp, cos, sin, tan, abs, trunc, sgn, round
    """

    kind: Literal["use_calculator_action"] = "use_calculator_action"
    expression: str = Field(description="math expression using previously known fact names and constants")
    fact_name: str = Field(
        description="fact name to save calculations result, should be unique, lowercase, snake_case, without spaces and special characters"
    )
    fact_unit: str = Field(description="expected unit of the fact value, if applicable, otherwise empty string")
    facts: dict | None = None


class CalculationResultObservation(Observation):
    kind: Literal["calculation_result_observation"] = "calculation_result_observation"
    name: str
    result: str


class Calculator(Tool):
    """
    Tool to evaluate math expressions
    """

    action: type[Action] = UseCalculatorAction
    observation: type[Observation] = CalculationResultObservation

    def execute_action(self, action: UseCalculatorAction) -> CalculationResultObservation:
        result = calculate(action.expression, action.facts or {})
        return CalculationResultObservation(name=action.fact_name, result=result)



import os
import sqlite3
import sys

_llm_calls = "LLMCalls"
_db = "tapedata.sqlite"


def add_rows_from_db(source_db_path, target_db_path):
    source_conn = sqlite3.connect(source_db_path)
    target_conn = sqlite3.connect(target_db_path)

    source_cursor = source_conn.cursor()
    target_cursor = target_conn.cursor()

    source_cursor.execute(f"SELECT * FROM {_llm_calls}")
    rows = source_cursor.fetchall()

    for row in rows:
        placeholders = ", ".join(["?"] * len(row))
        target_cursor.execute(f"INSERT INTO {_llm_calls} VALUES ({placeholders})", row)

    target_conn.commit()
    source_conn.close()
    target_conn.close()


def main():
    dirname = sys.argv[1]
    db_source = sys.argv[2]
    for root, dirs, files in os.walk(dirname):
        for file in files:
            if file.endswith(_db):
                db_path = os.path.join(root, file)
                print(db_path)
                add_rows_from_db(db_source, db_path)


if __name__ == "__main__":
    assert len(sys.argv) == 3, "Usage: python -m tests.add_sqlite_rows <dirname> <db_source>"
    main()



import logging
import multiprocessing
import os
import time

import hydra
from omegaconf import DictConfig, OmegaConf
from termcolor import colored

from examples.gaia_agent.steps import GaiaTape
from examples.rl_webagent.scripts.orchestrate_rl import load_webtasks
from tapeagents.core import Observation
from tapeagents.io import save_json_tape
from tapeagents.orchestrator import execute_agent, get_agent_and_env_from_config
from tapeagents.parallel_processing import process_pool_processor

logger = logging.getLogger(__name__)


def abt_to_json(tasks: list[dict]) -> list[dict]:
    """
    Convert AbstractBrowserTask tasks to dicts that are JSON serializable by only giving the dataset, task_id, and seed.
    """
    return [{"dataset": task["dataset"], "task": task["task"].get_task_id(), "seed": task["seed"]} for task in tasks]


def execute_single_task(cfg: DictConfig):
    """Worker function to execute a single task with its own agent and environment"""
    logging.basicConfig(
        format="%(asctime)s - PID_%(process)d - Thread_%(threadName)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
        handlers=[logging.StreamHandler()],
        force=True,  # forget previous handlers
    )
    logger.info(f"Processing task {cfg.task}")
    assert cfg.task, "Task must be provided in the configuration"

    try:
        agent, env = get_agent_and_env_from_config(cfg)
        tape, _ = env.start_task(cfg.task)  # type: ignore
        t = time.perf_counter()
        try:
            tape = execute_agent(agent, tape, env, max_loops=cfg.max_loops)
        except Exception as e:
            logger.error(f"Error occurred while running agent: {e}")
        logger.info("agent finished")
        tape.metadata.result = {"execution_time": time.perf_counter() - t}
        logger.info("save tape")
        save_json_tape(tape, os.path.join(cfg.exp_path, "tapes"), tape.metadata.parent_id)  # type: ignore
        logger.info("close environment")
        env.close()
    except Exception as e:
        logger.exception(f"Error during task execution: {e}")
        raise e
    return tape


@hydra.main(
    version_base=None,
    config_path=".",
    config_name="multiprocess_loop",
)
def main(cfg: DictConfig) -> None:
    os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(cfg.exp_path, "tapedata.sqlite")
    os.environ["MINIWOB_URL"] = cfg.environment_variables.miniwob_url

    train_samples, test_samples = load_webtasks(train_split=cfg.train_split, seeds=cfg.seeds)
    # train_samples, test_samples = load_webtasks_debug()  # TODO: load all tasks when ready
    # Convert AbstractBrowserTask to json parsable dicts
    train_samples = abt_to_json(train_samples)
    test_samples = abt_to_json(test_samples)

    task_args = []
    interpolated_cfg_dict = OmegaConf.to_container(cfg, resolve=True)
    for task in test_samples:
        task_cfg = OmegaConf.create(interpolated_cfg_dict)
        task_cfg.task = task
        task_args.append(task_cfg)

    logger.info(f"Processing {len(task_args)} tasks with {cfg.workers} workers")

    dt = time.perf_counter()

    results = []
    for result in process_pool_processor(
        stream=task_args,
        worker_func=execute_single_task,
        n_workers=cfg.workers,
        keep_order=False,
    ):
        if isinstance(result, Exception):
            logger.error(f"Task failed with exception: {result}")
            error_tape = GaiaTape(steps=[])
            error_tape.metadata.error = str(result)
            results.append(error_tape)
        else:
            results.append(result)
        logger.info(colored(f"{len(results)} of {len(task_args)} tasks processed so far", "blue"))

    ### Print some statistics
    total_steps = 0
    errs = 0
    acc = []
    rewards = []
    for tape in results:
        total_steps += len(tape.steps)
        last_obs = [step for step in tape if isinstance(step, Observation)][-1]
        success = last_obs.metadata.other.get("reward", 0.0) > 0.5
        acc.append(success)
        if (
            "info" in last_obs.metadata.other
            and "task_info" in last_obs.metadata.other["info"]
            and "REWARD_GLOBAL" in last_obs.metadata.other["info"]["task_info"]
        ):
            rewards.append(last_obs.metadata.other["info"]["task_info"]["REWARD_GLOBAL"])
        else:
            errs += 1
            logger.warning(f"No reward found in last observation of tape {tape.metadata.id}")

    logger.info(f"Average tape length: {(total_steps / len(results)):.2f}")
    logger.info(f"Total execution time: {time.perf_counter() - dt:.2f} seconds")
    logger.info(f"Average time per tape: {(time.perf_counter() - dt) / len(results):.2f} seconds")
    logger.info(f"Accuracy: {sum(acc) / len(acc) if acc else 0:.2f}")
    logger.info(f"Average reward: {sum(rewards) / len(rewards) if rewards else 0:.2f}")
    logger.info(f"Failed tapes: {errs}")


if __name__ == "__main__":
    # Set start method for multiprocessing (required for some systems)
    multiprocessing.set_start_method("spawn", force=True)
    main()
